"""Functions for reading and writing files."""
import json

from vigorish.enums import DataSet, LocalFileTask, S3FileTask, VigFile
from vigorish.util.sys_helpers import get_last_mod_time_utc
from vigorish.util.dt_format_strings import HTTP_TIME
from vigorish.util.result import Result
from vigorish.util.string_helpers import (
    validate_bbref_game_id,
    validate_brooks_game_id,
    validate_pitch_app_id,
)


class JsonStorage:
    """Perform CRUD operations on JSON files stored locally and/or in S3."""

    def __init__(self, config, file_helper):
        self.config = config
        self.file_helper = file_helper

    def save_json(self, data_set, parsed_data):
        local_filepath = None
        s3_object_key = None
        result_local = Result.Ok()
        result_s3 = Result.Ok()
        if self.json_stored_local_folder(VigFile.PARSED_JSON, data_set):
            result_local = self.save_json_local(data_set, parsed_data)
            if result_local.success:
                local_filepath = result_local.value
        if self.json_stored_s3(VigFile.PARSED_JSON, data_set):  # pragma: no cover
            result_s3 = self.save_json_s3(data_set, parsed_data)
            if result_s3.success:
                s3_object_key = result_s3.value
        result = Result.Combine([result_local, result_s3])
        if result.failure:
            return result
        return Result.Ok({"local_filepath": local_filepath, "s3_object_key": s3_object_key})

    def json_stored_local_folder(self, file_type, data_set):
        return self.file_helper.check_file_stored_local(file_type, data_set)

    def save_json_local(self, data_set, parsed_data):
        save_json_local_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.save_json_brooks_games_for_date_local_file,
            DataSet.BROOKS_PITCH_LOGS: self.save_json_brooks_pitch_logs_for_game_local_file,
            DataSet.BROOKS_PITCHFX: self.save_json_brooks_pitchfx_log_local_file,
            DataSet.BBREF_GAMES_FOR_DATE: self.save_json_bbref_games_for_date_local_file,
            DataSet.BBREF_BOXSCORES: self.save_json_bbref_boxscore_local_file,
        }
        return save_json_local_dict[data_set](parsed_data)

    def save_json_brooks_games_for_date_local_file(self, games_for_date):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=games_for_date.game_date,
            scraped_data=games_for_date,
        )

    def save_json_brooks_pitch_logs_for_game_local_file(self, pitch_logs_for_game):
        result = validate_brooks_game_id(pitch_logs_for_game.bb_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            scraped_data=pitch_logs_for_game,
            bb_game_id=pitch_logs_for_game.bb_game_id,
        )

    def save_json_brooks_pitchfx_log_local_file(self, pitchfx_log):
        result = validate_pitch_app_id(pitchfx_log.pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            scraped_data=pitchfx_log,
            pitch_app_id=pitchfx_log.pitch_app_id,
        )

    def save_json_bbref_games_for_date_local_file(self, games_for_date):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=games_for_date.game_date,
            scraped_data=games_for_date,
        )

    def save_json_bbref_boxscore_local_file(self, boxscore):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PARSED_JSON,
            game_date=boxscore.game_date,
            scraped_data=boxscore,
            bbref_game_id=boxscore.bbref_game_id,
        )

    def json_stored_s3(self, file_type, data_set):  # pragma: no cover
        return self.file_helper.check_file_stored_s3(file_type, data_set)

    def save_json_s3(self, data_set, parsed_data):  # pragma: no cover
        save_json_s3_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.save_json_brooks_games_for_date_s3,
            DataSet.BROOKS_PITCH_LOGS: self.save_json_brooks_pitch_logs_for_game_s3,
            DataSet.BROOKS_PITCHFX: self.save_json_brooks_pitchfx_log_s3,
            DataSet.BBREF_GAMES_FOR_DATE: self.save_json_bbref_games_for_date_s3,
            DataSet.BBREF_BOXSCORES: self.save_json_bbref_boxscore_s3,
        }
        return save_json_s3_dict[data_set](parsed_data)

    def save_json_brooks_games_for_date_s3(self, games_for_date):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=games_for_date.game_date,
            scraped_data=games_for_date,
        )

    def save_json_brooks_pitch_logs_for_game_s3(self, pitch_logs_for_game):  # pragma: no cover
        result = validate_brooks_game_id(pitch_logs_for_game.bb_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            scraped_data=pitch_logs_for_game,
            bb_game_id=pitch_logs_for_game.bb_game_id,
        )

    def save_json_brooks_pitchfx_log_s3(self, pitchfx_log):  # pragma: no cover
        result = validate_pitch_app_id(pitchfx_log.pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            scraped_data=pitchfx_log,
            pitch_app_id=pitchfx_log.pitch_app_id,
        )

    def save_json_bbref_games_for_date_s3(self, games_for_date):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=games_for_date.game_date,
            scraped_data=games_for_date,
        )

    def save_json_bbref_boxscore_s3(self, boxscore):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PARSED_JSON,
            game_date=boxscore.game_date,
            scraped_data=boxscore,
            bbref_game_id=boxscore.bbref_game_id,
        )

    def save_patch_list(self, data_set, patch_list):
        local_filepath = None
        s3_object_key = None
        result_local = Result.Ok()
        result_s3 = Result.Ok()
        if self.json_stored_local_folder(VigFile.PATCH_LIST, data_set):
            result_local = self.save_patch_list_local(data_set, patch_list)
            if result_local.success:
                local_filepath = result_local.value
        if self.json_stored_s3(VigFile.PATCH_LIST, data_set):  # pragma: no cover
            result_s3 = self.save_patch_list_s3(data_set, patch_list)
            if result_s3.success:
                s3_object_key = result_s3.value
        result = Result.Combine([result_local, result_s3])
        if result.failure:
            return result
        return Result.Ok({"local_filepath": local_filepath, "s3_object_key": s3_object_key})

    def save_patch_list_local(self, data_set, patch_list):
        save_patch_list_local_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.save_brooks_games_for_date_patch_list_local_file,
            DataSet.BROOKS_PITCHFX: self.save_brooks_pitchfx_patch_list_local_file,
            DataSet.BBREF_GAMES_FOR_DATE: self.save_bbref_games_for_date_patch_list_local_file,
            DataSet.BBREF_BOXSCORES: self.save_bbref_boxscore_patch_list_local_file,
        }
        save_patch_list_for_data_set = save_patch_list_local_dict.get(data_set)
        return save_patch_list_for_data_set(patch_list) if save_patch_list_for_data_set else None

    def save_brooks_games_for_date_patch_list_local_file(self, patch_list):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=patch_list.game_date,
            scraped_data=patch_list,
        )

    def save_brooks_pitchfx_patch_list_local_file(self, patch_list):
        result = validate_bbref_game_id(patch_list.url_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PATCH_LIST,
            game_date=game_dict["game_date"],
            scraped_data=patch_list,
            bbref_game_id=patch_list.url_id,
        )

    def save_bbref_games_for_date_patch_list_local_file(self, patch_list):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=patch_list.game_date,
            scraped_data=patch_list,
        )

    def save_bbref_boxscore_patch_list_local_file(self, patch_list):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PATCH_LIST,
            game_date=patch_list.game_date,
            scraped_data=patch_list,
            pitch_app_id=patch_list.url_id,
        )

    def save_patch_list_s3(self, data_set, patch_list):  # pragma: no cover
        save_patch_list_s3_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.save_brooks_games_for_date_patch_list_s3,
            DataSet.BROOKS_PITCHFX: self.save_brooks_pitchfx_patch_list_s3,
            DataSet.BBREF_GAMES_FOR_DATE: self.save_bbref_games_for_date_patch_list_s3,
            DataSet.BBREF_BOXSCORES: self.save_bbref_boxscore_patch_list_s3,
        }
        save_patch_list_for_data_set = save_patch_list_s3_dict.get(data_set)
        return save_patch_list_for_data_set(patch_list) if save_patch_list_for_data_set else None

    def save_brooks_games_for_date_patch_list_s3(self, patch_list):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=patch_list.game_date,
            scraped_data=patch_list,
        )

    def save_brooks_pitchfx_patch_list_s3(self, patch_list):  # pragma: no cover
        result = validate_bbref_game_id(patch_list.url_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PATCH_LIST,
            game_date=game_dict["game_date"],
            scraped_data=patch_list,
            bbref_game_id=patch_list.url_id,
        )

    def save_bbref_games_for_date_patch_list_s3(self, patch_list):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=patch_list.game_date,
            scraped_data=patch_list,
        )

    def save_bbref_boxscore_patch_list_s3(self, patch_list):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PATCH_LIST,
            game_date=patch_list.game_date,
            scraped_data=patch_list,
            pitch_app_id=patch_list.url_id,
        )

    def save_combined_game_data(self, combined_data):
        local_filepath = None
        s3_object_key = None
        result_local = Result.Ok()
        result_s3 = Result.Ok()
        if self.json_stored_local_folder(VigFile.COMBINED_GAME_DATA, DataSet.ALL):
            result_local = self.save_combined_game_data_local_file(combined_data)
            if result_local.success:
                local_filepath = result_local.value
        if self.json_stored_s3(VigFile.COMBINED_GAME_DATA, DataSet.ALL):  # pragma: no cover
            result_s3 = self.save_combined_game_data_s3(combined_data)
            if result_s3.success:
                s3_object_key = result_s3.value
        result = Result.Combine([result_local, result_s3])
        if result.failure:
            return result
        return Result.Ok({"local_filepath": local_filepath, "s3_object_key": s3_object_key})

    def save_combined_game_data_local_file(self, combined_data):
        result = validate_bbref_game_id(combined_data["bbref_game_id"])
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.WRITE_FILE,
            data_set=DataSet.ALL,
            file_type=VigFile.COMBINED_GAME_DATA,
            game_date=game_dict["game_date"],
            scraped_data=combined_data,
            bbref_game_id=combined_data["bbref_game_id"],
        )

    def save_combined_game_data_s3(self, combined_data):  # pragma: no cover
        result = validate_bbref_game_id(combined_data["bbref_game_id"])
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.UPLOAD,
            data_set=DataSet.ALL,
            file_type=VigFile.COMBINED_GAME_DATA,
            game_date=game_dict["game_date"],
            scraped_data=combined_data,
            bbref_game_id=combined_data["bbref_game_id"],
        )

    def get_combined_game_data(self, bbref_game_id):
        if self.json_stored_local_folder(VigFile.COMBINED_GAME_DATA, DataSet.ALL):
            result = self.decode_combined_game_data_local_file(bbref_game_id)
            if result.success:
                return result.value
        if self.json_stored_s3(VigFile.COMBINED_GAME_DATA, DataSet.ALL):  # pragma: no cover
            result = self.decode_combined_game_data_s3(bbref_game_id)
            if result.success:
                return result.value
        return None

    def decode_combined_game_data_local_file(self, bbref_game_id):
        result = self.get_combined_game_data_local_file(bbref_game_id)
        if result.failure:
            return result
        filepath = result.value
        try:
            json_dict = json.loads(filepath.read_text())
            json_dict["last_modified"] = get_last_mod_time_utc(filepath).strftime(HTTP_TIME)
            return Result.Ok(json_dict)
        except Exception as e:
            error = f"Error: {repr(e)}"
            return Result.Fail(error)

    def get_combined_game_data_local_file(self, bbref_game_id):
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.ALL,
            file_type=VigFile.COMBINED_GAME_DATA,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def decode_combined_game_data_s3(self, bbref_game_id):  # pragma: no cover
        result = self.get_combined_game_data_s3(bbref_game_id)
        if result.failure:
            return result
        filepath = result.value
        try:
            json_dict = json.loads(filepath.read_text())
            json_dict["last_modified"] = get_last_mod_time_utc(filepath).strftime(HTTP_TIME)
            return Result.Ok(json_dict)
        except Exception as e:
            error = f"Error: {repr(e)}"
            return Result.Fail(error)

    def get_combined_game_data_s3(self, bbref_game_id):  # pragma: no cover
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.ALL,
            file_type=VigFile.COMBINED_GAME_DATA,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def get_scraped_data(self, data_set, url_id):
        if self.json_stored_local_folder(VigFile.PARSED_JSON, data_set):
            result = self.get_scraped_data_local(data_set, url_id)
            if result.success:
                return result.value
        if self.json_stored_s3(VigFile.PARSED_JSON, data_set):  # pragma: no cover
            result = self.get_scraped_data_s3(data_set, url_id)
            if result.success:
                return result.value
        return None

    def get_scraped_data_local(self, data_set, url_id):
        get_scraped_data_local_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.decode_json_brooks_games_for_date_local_file,
            DataSet.BROOKS_PITCH_LOGS: self.decode_json_brooks_pitch_logs_for_game_local_file,
            DataSet.BROOKS_PITCHFX: self.decode_json_brooks_pitchfx_log_local_file,
            DataSet.BBREF_GAMES_FOR_DATE: self.decode_json_bbref_games_for_date_local_file,
            DataSet.BBREF_BOXSCORES: self.decode_json_bbref_boxscore_local_file,
        }
        return get_scraped_data_local_dict[data_set](url_id)

    def decode_json_brooks_games_for_date_local_file(self, game_date):
        result = self.get_json_brooks_games_for_date_local_file(game_date)
        if result.failure:
            return result
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
            delete_file=False,
        )

    def get_json_brooks_games_for_date_local_file(self, game_date):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
        )

    def decode_json_brooks_pitch_logs_for_game_local_file(self, bb_game_id):
        result = self.get_json_brooks_pitch_logs_for_game_local_file(bb_game_id)
        if result.failure:
            return result
        result = validate_brooks_game_id(bb_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bb_game_id=bb_game_id,
            delete_file=False,
        )

    def get_json_brooks_pitch_logs_for_game_local_file(self, bb_game_id):
        result = validate_brooks_game_id(bb_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bb_game_id=bb_game_id,
        )

    def decode_json_brooks_pitchfx_log_local_file(self, pitch_app_id):
        result = self.get_json_brooks_pitchfx_local_file(pitch_app_id)
        if result.failure:
            return result
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
            delete_file=False,
        )

    def get_json_brooks_pitchfx_local_file(self, pitch_app_id):
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def decode_json_bbref_games_for_date_local_file(self, game_date):
        result = self.get_json_bbref_games_for_date_local_file(game_date)
        if result.failure:
            return result
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
            delete_file=False,
        )

    def get_json_bbref_games_for_date_local_file(self, game_date):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
        )

    def decode_json_bbref_boxscore_local_file(self, bbref_game_id):
        result = self.get_json_bbref_boxscore_local_file(bbref_game_id)
        if result.failure:
            return result
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
            delete_file=False,
        )

    def get_json_bbref_boxscore_local_file(self, bbref_game_id):
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def get_scraped_data_s3(self, data_set, url_id):  # pragma: no cover
        get_scraped_data_s3_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.decode_json_brooks_games_for_date_s3,
            DataSet.BROOKS_PITCH_LOGS: self.decode_json_brooks_pitch_logs_for_game_s3,
            DataSet.BROOKS_PITCHFX: self.decode_json_brooks_pitchfx_log_s3,
            DataSet.BBREF_GAMES_FOR_DATE: self.decode_json_bbref_games_for_date_s3,
            DataSet.BBREF_BOXSCORES: self.decode_json_bbref_boxscore_s3,
        }
        return get_scraped_data_s3_dict[data_set](url_id)

    def decode_json_brooks_games_for_date_s3(self, game_date):  # pragma: no cover
        result = self.get_json_brooks_games_for_date_s3(game_date)
        if result.failure:
            return result
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
            delete_file=True,
        )

    def get_json_brooks_games_for_date_s3(self, game_date):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
        )

    def decode_json_brooks_pitchfx_log_s3(self, pitch_app_id):  # pragma: no cover
        result = self.get_json_brooks_pitchfx_s3(pitch_app_id)
        if result.failure:
            return result
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
            delete_file=True,
        )

    def decode_json_brooks_pitch_logs_for_game_s3(self, bb_game_id):  # pragma: no cover
        result = self.get_json_brooks_pitch_logs_for_game_s3(bb_game_id)
        if result.failure:
            return result
        result = validate_brooks_game_id(bb_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bb_game_id=bb_game_id,
            delete_file=True,
        )

    def get_json_brooks_pitch_logs_for_game_s3(self, bb_game_id):  # pragma: no cover
        result = validate_brooks_game_id(bb_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bb_game_id=bb_game_id,
        )

    def get_json_brooks_pitchfx_s3(self, pitch_app_id):  # pragma: no cover
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def decode_json_bbref_games_for_date_s3(self, game_date):  # pragma: no cover
        result = self.get_json_bbref_games_for_date_s3(game_date)
        if result.failure:
            return result
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
            delete_file=True,
        )

    def get_json_bbref_games_for_date_s3(self, game_date):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
        )

    def decode_json_bbref_boxscore_s3(self, bbref_game_id):  # pragma: no cover
        result = self.get_json_bbref_boxscore_s3(bbref_game_id)
        if result.failure:
            return result
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
            delete_file=True,
        )

    def get_json_bbref_boxscore_s3(self, bbref_game_id):  # pragma: no cover
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def get_patch_list(self, data_set, url_id):
        if self.json_stored_local_folder(VigFile.PATCH_LIST, data_set):
            result = self.get_patch_list_local(data_set, url_id)
            if result.success:
                return result.value
        if self.json_stored_s3(VigFile.PATCH_LIST, data_set):  # pragma: no cover
            result = self.get_patch_list_s3(data_set, url_id)
            if result.success:
                return result.value
        return None

    def get_patch_list_local(self, data_set, url_id):
        get_patch_list_local_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.decode_brooks_games_for_date_patch_list_local_file,
            DataSet.BROOKS_PITCHFX: self.decode_brooks_pitchfx_patch_list_local_file,
            DataSet.BBREF_GAMES_FOR_DATE: self.decode_bbref_games_for_date_patch_list_local_file,
            DataSet.BBREF_BOXSCORES: self.decode_bbref_boxscore_patch_list_local_file,
        }
        get_patch_list_for_data_set = get_patch_list_local_dict.get(data_set)
        return get_patch_list_for_data_set(url_id) if get_patch_list_for_data_set else None

    def decode_brooks_games_for_date_patch_list_local_file(self, game_date):
        result = self.get_brooks_games_for_date_patch_list_local_file(game_date)
        if result.failure:
            return result
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=game_date,
            delete_file=False,
        )

    def get_brooks_games_for_date_patch_list_local_file(self, game_date):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=game_date,
        )

    def decode_brooks_pitchfx_patch_list_local_file(self, bbref_game_id):
        result = self.get_brooks_pitchfx_patch_list_local_file(bbref_game_id)
        if result.failure:
            return result
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PATCH_LIST,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
            delete_file=False,
        )

    def get_brooks_pitchfx_patch_list_local_file(self, bbref_game_id):
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PATCH_LIST,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def decode_bbref_games_for_date_patch_list_local_file(self, game_date):
        result = self.get_bbref_games_for_date_patch_list_local_file(game_date)
        if result.failure:
            return result
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=game_date,
            delete_file=False,
        )

    def get_bbref_games_for_date_patch_list_local_file(self, game_date):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=game_date,
        )

    def decode_bbref_boxscore_patch_list_local_file(self, bbref_game_id):
        result = self.get_bbref_boxscore_patch_list_local_file(bbref_game_id)
        if result.failure:
            return result
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PATCH_LIST,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
            delete_file=False,
        )

    def get_bbref_boxscore_patch_list_local_file(self, bbref_game_id):
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.READ_FILE,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PATCH_LIST,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def get_patch_list_s3(self, data_set, url_id):  # pragma: no cover
        get_patch_list_s3_dict = {
            DataSet.BROOKS_GAMES_FOR_DATE: self.decode_brooks_games_for_date_patch_list_s3,
            DataSet.BROOKS_PITCHFX: self.decode_brooks_pitchfx_patch_list_s3,
            DataSet.BBREF_GAMES_FOR_DATE: self.decode_bbref_games_for_date_patch_list_s3,
            DataSet.BBREF_BOXSCORES: self.decode_bbref_boxscore_patch_list_s3,
        }
        get_patch_list_for_data_set = get_patch_list_s3_dict.get(data_set)
        return get_patch_list_for_data_set(url_id) if get_patch_list_for_data_set else None

    def decode_brooks_games_for_date_patch_list_s3(self, game_date):  # pragma: no cover
        result = self.get_brooks_games_for_date_patch_list_s3(game_date)
        if result.failure:
            return result
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=game_date,
            delete_file=True,
        )

    def get_brooks_games_for_date_patch_list_s3(self, game_date):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=game_date,
        )

    def decode_brooks_pitchfx_patch_list_s3(self, bbref_game_id):  # pragma: no cover
        result = self.get_brooks_pitchfx_patch_list_local_file(bbref_game_id)
        if result.failure:
            return result
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PATCH_LIST,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
            delete_file=True,
        )

    def get_brooks_pitchfx_patch_list_s3(self, bbref_game_id):  # pragma: no cover
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PATCH_LIST,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def decode_bbref_games_for_date_patch_list_s3(self, game_date):  # pragma: no cover
        result = self.get_bbref_games_for_date_patch_list_s3(game_date)
        if result.failure:
            return result
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=game_date,
            delete_file=True,
        )

    def get_bbref_games_for_date_patch_list_s3(self, game_date):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PATCH_LIST,
            game_date=game_date,
        )

    def decode_bbref_boxscore_patch_list_s3(self, bbref_game_id):  # pragma: no cover
        result = self.get_bbref_boxscore_patch_list_s3(bbref_game_id)
        if result.failure:
            return result
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DECODE_JSON,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PATCH_LIST,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
            delete_file=True,
        )

    def get_bbref_boxscore_patch_list_s3(self, bbref_game_id):  # pragma: no cover
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DOWNLOAD,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PATCH_LIST,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def delete_brooks_games_for_date_local_file(self, game_date):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DELETE_FILE,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
        )

    def delete_brooks_pitch_logs_for_game_local_file(self, bb_game_id):
        result = validate_brooks_game_id(bb_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DELETE_FILE,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bb_game_id=bb_game_id,
        )

    def delete_brooks_pitchfx_log_local_file(self, pitch_app_id):
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DELETE_FILE,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def delete_bbref_games_for_date_local_file(self, game_date):
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DELETE_FILE,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
        )

    def delete_bbref_boxscore_local_file(self, bbref_game_id):
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_local_file_task(
            task=LocalFileTask.DELETE_FILE,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def delete_brooks_games_for_date_s3(self, game_date):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DELETE,
            data_set=DataSet.BROOKS_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
        )

    def delete_brooks_pitch_logs_for_game_s3(self, bb_game_id):  # pragma: no cover
        result = validate_brooks_game_id(bb_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DELETE,
            data_set=DataSet.BROOKS_PITCH_LOGS,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bb_game_id=bb_game_id,
        )

    def delete_brooks_pitchfx_log_s3(self, pitch_app_id):  # pragma: no cover
        result = validate_pitch_app_id(pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DELETE,
            data_set=DataSet.BROOKS_PITCHFX,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            pitch_app_id=pitch_app_id,
        )

    def delete_bbref_games_for_date_s3(self, game_date):  # pragma: no cover
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DELETE,
            data_set=DataSet.BBREF_GAMES_FOR_DATE,
            file_type=VigFile.PARSED_JSON,
            game_date=game_date,
        )

    def delete_bbref_boxscore_s3(self, bbref_game_id):  # pragma: no cover
        result = validate_bbref_game_id(bbref_game_id)
        if result.failure:
            return result
        game_dict = result.value
        return self.file_helper.perform_s3_task(
            task=S3FileTask.DELETE,
            data_set=DataSet.BBREF_BOXSCORES,
            file_type=VigFile.PARSED_JSON,
            game_date=game_dict["game_date"],
            bbref_game_id=bbref_game_id,
        )

    def rename_brooks_pitchfx_log(self, old_pitch_app_id, new_pitch_app_id, year):  # pragma: no cover
        result = validate_pitch_app_id(old_pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        old_key = self.file_helper.get_object_key(
            file_type=VigFile.PARSED_JSON,
            data_set=DataSet.BROOKS_PITCHFX,
            game_date=game_dict["game_date"],
            pitch_app_id=old_pitch_app_id,
        )
        result = validate_pitch_app_id(new_pitch_app_id)
        if result.failure:
            return result
        game_dict = result.value
        new_key = self.file_helper.get_object_key(
            file_type=VigFile.PARSED_JSON,
            data_set=DataSet.BROOKS_PITCHFX,
            game_date=game_dict["game_date"],
            pitch_app_id=new_pitch_app_id,
        )
        return self.file_helper.rename_s3_object(old_key, new_key)



from dataclasses import dataclass

from dataclass_csv import accept_whitespaces
from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

import vigorish.database as db


class PitchAppScrapeStatus(db.Base):
    __tablename__ = "scrape_status_pitch_app"
    id = Column(Integer, primary_key=True)
    pitcher_id_mlb = Column(Integer)
    pitch_app_id = Column(String, unique=True)
    bbref_game_id = Column(String)
    bb_game_id = Column(String)
    scraped_pitchfx = Column(Integer, default=0)
    no_pitchfx_data = Column(Integer, default=0)
    combined_pitchfx_bbref_data = Column(Integer, default=0)
    imported_pitchfx = Column(Integer, default=0)
    invalid_pitchfx = Column(Integer, default=0)
    pitchfx_error = Column(Integer, default=0)
    pitch_count_pitch_log = Column(Integer, default=0)
    pitch_count_bbref = Column(Integer, default=0)
    pitch_count_pitchfx = Column(Integer, default=0)
    pitch_count_pitchfx_audited = Column(Integer, default=0)
    batters_faced_bbref = Column(Integer, default=0)
    batters_faced_pitchfx = Column(Integer, default=0)
    patched_pitchfx_count = Column(Integer, default=0)
    missing_pitchfx_count = Column(Integer, default=0)
    removed_pitchfx_count = Column(Integer, default=0)
    invalid_pitchfx_count = Column(Integer, default=0)
    total_at_bats_pitchfx_complete = Column(Integer, default=0)
    total_at_bats_patched_pitchfx = Column(Integer, default=0)
    total_at_bats_missing_pitchfx = Column(Integer, default=0)
    total_at_bats_removed_pitchfx = Column(Integer, default=0)
    total_at_bats_pitchfx_error = Column(Integer, default=0)
    total_at_bats_invalid_pitchfx = Column(Integer, default=0)

    pitcher_id = Column(Integer, ForeignKey("player.id"), index=True)
    scrape_status_game_id = Column(Integer, ForeignKey("scrape_status_game.id"), index=True)
    scrape_status_date_id = Column(Integer, ForeignKey("scrape_status_date.id"))
    season_id = Column(Integer, ForeignKey("season.id"))

    pitchfx = relationship("PitchFx")

    @hybrid_property
    def game_date(self):
        return self.date.game_date

    @hybrid_property
    def contains_only_patched_data(self):
        return self.patched_pitchfx_count > 0 and (self.pitch_count_pitchfx_audited == self.patched_pitchfx_count)

    def __repr__(self):
        return f'<PitchAppScrapeStatus pitch_app_id="{self.pitch_app_id}">'

    @classmethod
    def find_by_pitch_app_id(cls, db_session, pitch_app_id):
        return db_session.query(cls).filter_by(pitch_app_id=pitch_app_id).first()

    @classmethod
    def get_all_scraped_pitch_app_ids_for_season(cls, db_session, season_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if pitch_app_status.scraped_pitchfx == 1
        ]

    @classmethod
    def get_all_pitch_app_ids_for_game(cls, db_session, bbref_game_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls).filter_by(bbref_game_id=bbref_game_id).all()
        ]

    @classmethod
    def get_all_pitch_app_ids_with_pfx_data_for_game(cls, db_session, bbref_game_id):
        pitch_apps_with_pfx_data = list(
            db_session.query(cls)
            .filter_by(bbref_game_id=bbref_game_id)
            .filter_by(scraped_pitchfx=1)
            .filter_by(no_pitchfx_data=0)
            .all()
        )

        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in pitch_apps_with_pfx_data
            if not pitch_app_status.contains_only_patched_data
        ]

    @classmethod
    def get_all_unscraped_pitch_app_ids_for_game(cls, db_session, bbref_game_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls)
            .filter_by(bbref_game_id=bbref_game_id)
            .filter_by(scraped_pitchfx=0)
            .all()
        ]

    @classmethod
    def get_all_scraped_pitch_app_ids_for_game(cls, db_session, bbref_game_id):
        return [
            pitch_app_status.pitch_app_id
            for pitch_app_status in db_session.query(cls)
            .filter_by(bbref_game_id=bbref_game_id)
            .filter_by(scraped_pitchfx=1)
            .all()
        ]

    @classmethod
    def get_pitch_app_id_map(cls, db_session):
        all_pitch_apps = db_session.query(cls).all()
        return {pa.pitch_app_id: pa.id for pa in all_pitch_apps}


@accept_whitespaces
@dataclass
class PitchAppScrapeStatusCsvRow:
    id: int
    pitcher_id_mlb: int = 0
    pitch_app_id: str = None
    bbref_game_id: str = None
    bb_game_id: str = None
    scraped_pitchfx: int = 0
    no_pitchfx_data: int = 0
    combined_pitchfx_bbref_data: int = 0
    imported_pitchfx: int = 0
    invalid_pitchfx: int = 0
    pitchfx_error: int = 0
    pitch_count_pitch_log: int = 0
    pitch_count_bbref: int = 0
    pitch_count_pitchfx: int = 0
    pitch_count_pitchfx_audited: int = 0
    batters_faced_bbref: int = 0
    batters_faced_pitchfx: int = 0
    patched_pitchfx_count: int = 0
    missing_pitchfx_count: int = 0
    removed_pitchfx_count: int = 0
    invalid_pitchfx_count: int = 0
    total_at_bats_pitchfx_complete: int = 0
    total_at_bats_patched_pitchfx: int = 0
    total_at_bats_missing_pitchfx: int = 0
    total_at_bats_removed_pitchfx: int = 0
    total_at_bats_pitchfx_error: int = 0
    total_at_bats_invalid_pitchfx: int = 0



import errno
import json
import os
from copy import deepcopy
from pathlib import Path

from vigorish.config.config_setting import (
    EnumConfigSetting,
    NumericConfigSetting,
    PathConfigSetting,
    StringConfigSetting,
)
from vigorish.config.types.batch_job_settings import BatchJobSettings
from vigorish.config.types.batch_scrape_delay import BatchScrapeDelay
from vigorish.config.types.url_scrape_delay import UrlScrapeDelay
from vigorish.enums import (
    CombinedDataStorageOption,
    DataSet,
    HtmlStorageOption,
    JsonStorageOption,
    ScrapeCondition,
    ScrapeTaskOption,
    StatusReport,
)
from vigorish.util.list_helpers import dict_to_param_list
from vigorish.util.result import Result

VIG_FOLDER = Path.home() / ".vig"
DEFAULT_CONFIG = VIG_FOLDER / "vig.config.json"

# TODO: Create new config setting: DB_TABLE_BATCH_SIZE
#       Used with add_to_database and restore_database tasks


class ConfigFile:
    config_filepath: str

    def __init__(self, config_file_path=None):
        self.config_filepath = config_file_path
        if not self.config_filepath:
            self.config_filepath = Path(os.getenv("CONFIG_FILE"))
        if not self.config_filepath:
            self.config_filepath = DEFAULT_CONFIG
        if isinstance(self.config_filepath, str):
            self.config_filepath = Path(self.config_filepath)
        if self.config_filepath.exists():
            self.read_config_file()
        else:
            self.create_default_config_file()

    @property
    def all_settings(self):
        return {name: self.config_factory(name, config) for name, config in self.config_json.items()}

    def get_current_setting(self, setting_name, data_set=DataSet.ALL, year=None):
        config_setting = self.all_settings.get(setting_name)
        if not config_setting:
            raise ValueError(f"{setting_name} is not a valid configuration setting")
        return (
            config_setting.current_setting(data_set).resolve(year)
            if isinstance(config_setting, PathConfigSetting)
            else config_setting.current_setting(data_set)
        )

    def change_setting(self, setting_name, data_set, new_value):
        config_dict = self.config_json.get(setting_name)
        if config_dict:
            config_dict["SAME_SETTING_FOR_ALL_DATA_SETS"] = data_set == DataSet.ALL
            if config_dict["CONFIG_TYPE"] == "Numeric":
                result = self.get_object(setting_name, new_value)
                if result.failure:
                    return result
                config_dict[data_set.name] = result.value
                self.reset_other_data_sets_numeric(config_dict, data_set)
            else:
                config_dict[data_set.name] = str(new_value)
                self.reset_other_data_sets_enum_str(config_dict, data_set)
        return self.write_config_file()

    def get_all_url_scrape_params(self, data_set):
        url_delay_settings = self.get_current_setting("URL_SCRAPE_DELAY", data_set)
        batch_job_settings = self.get_current_setting("BATCH_JOB_SETTINGS", data_set)
        batch_delay_settings = self.get_current_setting("BATCH_SCRAPE_DELAY", data_set)
        if url_delay_settings and batch_job_settings and batch_delay_settings:
            return self.get_nodejs_script_params_from_objects(
                url_delay_settings, batch_job_settings, batch_delay_settings
            )
        else:
            return self.get_default_nodejs_script_params()

    def get_nodejs_script_args(self, data_set, url_set_filepath):
        script_params = self.get_all_url_scrape_params(data_set)
        script_params["urlSetFilepath"] = url_set_filepath.resolve()
        return dict_to_param_list(script_params)

    def check_url_delay_settings(self, data_sets):
        url_delay_settings = [self.get_current_setting("URL_SCRAPE_DELAY", data_set) for data_set in data_sets]
        if all(url_delay.is_valid for url_delay in url_delay_settings):
            return Result.Ok()
        error = "URL delay cannot be disabled and must be at least 3 seconds"
        return Result.Fail(error)

    def s3_bucket_required(self, data_sets):
        html_storage_settings = [self.get_current_setting("HTML_STORAGE", data_set) for data_set in data_sets]
        json_storage_settings = [self.get_current_setting("JSON_STORAGE", data_set) for data_set in data_sets]
        html_local_storage = all(
            html_storage in [HtmlStorageOption.NONE, HtmlStorageOption.LOCAL_FOLDER]
            for html_storage in html_storage_settings
        )

        json_local_storage = all(
            json_storage == JsonStorageOption.LOCAL_FOLDER for json_storage in json_storage_settings
        )
        return not html_local_storage or not json_local_storage

    def read_config_file(self):
        if not self.config_filepath.exists():
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), str(self.config_filepath))
        if not self.config_filepath.is_file():
            raise TypeError(f"Unable to open config file: {self.config_filepath}")
        self.config_json = json.loads(self.config_filepath.read_text())

    def write_config_file(self):
        try:
            config_json = json.dumps(self.config_json, indent=2, sort_keys=False)
            self.config_filepath.write_text(config_json)
            return Result.Ok()
        except Exception as e:
            error = f"Error: {repr(e)}"
            return Result.Fail(error)

    def config_factory(self, setting_name, config_dict):
        config_type = config_dict.get("CONFIG_TYPE")
        if config_type == "Enum":
            return EnumConfigSetting(setting_name, config_dict)
        if config_type == "Numeric":
            return NumericConfigSetting(setting_name, config_dict)
        if config_type == "Path":
            return PathConfigSetting(setting_name, config_dict)
        return StringConfigSetting(setting_name, config_dict)

    def reset_other_data_sets_enum_str(self, setting, data_set):
        if data_set == DataSet.ALL:
            setting[DataSet.BBREF_BOXSCORES.name] = None
            setting[DataSet.BBREF_GAMES_FOR_DATE.name] = None
            setting[DataSet.BROOKS_GAMES_FOR_DATE.name] = None
            setting[DataSet.BROOKS_PITCHFX.name] = None
            setting[DataSet.BROOKS_PITCH_LOGS.name] = None
        else:
            setting[DataSet.ALL.name] = None

    def reset_other_data_sets_numeric(self, setting, data_set):
        class_name = setting.get("CLASS_NAME")
        null_object = self.get_null_object(class_name)
        if data_set == DataSet.ALL:
            setting[DataSet.BBREF_BOXSCORES.name] = null_object
            setting[DataSet.BBREF_GAMES_FOR_DATE.name] = null_object
            setting[DataSet.BROOKS_GAMES_FOR_DATE.name] = null_object
            setting[DataSet.BROOKS_PITCHFX.name] = null_object
            setting[DataSet.BROOKS_PITCH_LOGS.name] = null_object
        else:
            setting[DataSet.ALL.name] = null_object

    def get_object(self, setting_name, new_value):
        setting = self.config_json.get(setting_name)
        class_name = setting.get("CLASS_NAME")
        if class_name == "UrlScrapeDelay":
            result = self.validate_new_url_delay_setting(new_value)
            if result.failure:
                return result
            return Result.Ok(UrlScrapeDelay(*new_value).to_dict())
        if class_name == "BatchJobSettings":
            return Result.Ok(BatchJobSettings(*new_value).to_dict())
        if class_name == "BatchScrapeDelay":
            return Result.Ok(BatchScrapeDelay(*new_value).to_dict())

    def validate_new_url_delay_setting(self, new_value):
        is_enabled, is_random, delay_uniform, delay_min, delay_max = new_value
        if not is_enabled:
            return Result.Fail("URL delay cannot be disabled!")
        if not is_random and delay_uniform < 3 or is_random and delay_min < 3:
            return Result.Fail("URL delay min value must be greater than 2 seconds!")
        return Result.Ok()

    def get_null_object(self, class_name):
        if class_name == "UrlScrapeDelay":
            return UrlScrapeDelay.null_object()
        if class_name == "BatchJobSettings":
            return BatchJobSettings.null_object()
        if class_name == "BatchScrapeDelay":
            return BatchScrapeDelay.null_object()

    def get_default_nodejs_script_params(self):
        return {
            "urlTimeoutRequired": True,
            "urlTimeoutMinMs": 3000,
            "urlTimeoutMaxMs": 6000,
            "batchTimeoutRequired": True,
            "batchTimeoutMinMs": 1800000,
            "batchTimeoutMaxMs": 2700000,
            "batchScrapingEnabled": True,
            "minBatchSize": 50,
            "maxBatchSize": 80,
        }

    def get_nodejs_script_params_from_objects(self, url_delay_settings, batch_job_settings, batch_delay_settings):
        script_params = {}
        if url_delay_settings.delay_is_required:
            script_params["urlTimeoutRequired"] = True
            if url_delay_settings.delay_is_random:
                script_params["urlTimeoutMinMs"] = url_delay_settings.delay_random_min_ms
                script_params["urlTimeoutMaxMs"] = url_delay_settings.delay_random_max_ms
            else:
                script_params["urlTimeoutUniformMs"] = url_delay_settings.delay_uniform_ms
        if batch_delay_settings.delay_is_required:
            script_params["batchTimeoutRequired"] = True
            if batch_delay_settings.delay_is_random:
                script_params["batchTimeoutMinMs"] = batch_delay_settings.delay_random_min_ms
                script_params["batchTimeoutMaxMs"] = batch_delay_settings.delay_random_max_ms
            else:
                script_params["batchTimeoutUniformMs"] = batch_delay_settings.delay_uniform_ms
        if batch_job_settings.batched_scraping_enabled:
            script_params["batchScrapingEnabled"] = True
            if batch_job_settings.batch_size_is_random:
                script_params["minBatchSize"] = batch_job_settings.batch_size_random_min
                script_params["maxBatchSize"] = batch_job_settings.batch_size_random_max
            else:
                script_params["uniformBatchSize"] = batch_job_settings.batch_size_uniform
        return script_params

    def create_default_config_file(self):
        self.config_json = deepcopy(self.settings_meta)
        for setting, config_dict in self.config_json.items():
            if config_dict["SAME_SETTING_FOR_ALL_DATA_SETS"]:
                config_dict[DataSet.ALL.name] = self.get_default_value(setting, DataSet.ALL)
                for data_set in DataSet:
                    config_dict[data_set.name] = None
            else:
                config_dict[DataSet.ALL.name] = None
                for data_set in DataSet:
                    config_dict[data_set.name] = self.get_default_value(setting, data_set)
        self.write_config_file()

    def get_default_value(self, setting_name, data_set):
        default_value_dict = {
            "STATUS_REPORT": StatusReport.SEASON_SUMMARY.name,
            "S3_BUCKET": "your-bucket",
            "SCRAPE_CONDITION": ScrapeCondition.ONLY_MISSING_DATA.name,
            "SCRAPE_TASK_OPTION": ScrapeTaskOption.BY_DATE.name,
            "URL_SCRAPE_DELAY": UrlScrapeDelay(True, True, None, 3, 6).to_dict(),
            "BATCH_JOB_SETTINGS": BatchJobSettings(True, True, None, 50, 80).to_dict(),
            "HTML_STORAGE": HtmlStorageOption.NONE.name,
            "HTML_LOCAL_FOLDER_PATH": "html_storage/{year}/{data_set}/",
            "HTML_S3_FOLDER_PATH": "{year}/{data_set}/html/",
            "JSON_STORAGE": JsonStorageOption.LOCAL_FOLDER.name,
            "JSON_LOCAL_FOLDER_PATH": "json_storage/{year}/{data_set}/",
            "JSON_S3_FOLDER_PATH": "{year}/{data_set}/",
            "SCRAPED_DATA_COMBINE_CONDITION": ScrapeCondition.ONLY_MISSING_DATA.name,
            "COMBINED_DATA_STORAGE": CombinedDataStorageOption.LOCAL_FOLDER.name,
            "COMBINED_DATA_LOCAL_FOLDER_PATH": "json_storage/{year}/combined_data",
            "COMBINED_DATA_S3_FOLDER_PATH": "{year}/combined_data",
            "DB_BACKUP_FOLDER_PATH": "backup",
        }
        batch_delay_setting_dict = {
            "BBREF_GAMES_FOR_DATE": BatchScrapeDelay(True, True, None, 5, 10).to_dict(),
            "BBREF_BOXSCORES": BatchScrapeDelay(True, True, None, 5, 10).to_dict(),
            "BROOKS_GAMES_FOR_DATE": BatchScrapeDelay(True, True, None, 30, 45).to_dict(),
            "BROOKS_PITCH_LOGS": BatchScrapeDelay(True, True, None, 30, 45).to_dict(),
            "BROOKS_PITCHFX": BatchScrapeDelay(True, True, None, 30, 45).to_dict(),
        }
        if setting_name in default_value_dict:
            return default_value_dict.get(setting_name)
        if setting_name == "BATCH_SCRAPE_DELAY":
            return batch_delay_setting_dict.get(data_set.name)
        raise ValueError(f"{setting_name} is not a valid setting name")

    @property
    def settings_meta(self):
        return {
            "STATUS_REPORT": {
                "CONFIG_TYPE": "Enum",
                "ENUM_NAME": "StatusReport",
                "DESCRIPTION": (
                    "After a scrape job has successfully completed, you can display a report "
                    "for the MLB season displaying various metrics for all data sets. The "
                    "options below determine the level of detail reported."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "S3_BUCKET": {
                "CONFIG_TYPE": "str",
                "DESCRIPTION": ("The name of the S3 bucket to use for storing HTML and/or JSON files"),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "SCRAPE_CONDITION": {
                "CONFIG_TYPE": "Enum",
                "ENUM_NAME": "ScrapeCondition",
                "DESCRIPTION": (
                    "By default, HTML is scraped and parsed only once (ONLY_MISSING_DATA). "
                    "You can overwrite existing data by selecting ALWAYS, or prevent any "
                    "data from being scraped by selecting NEVER."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "SCRAPE_TASK_OPTION": {
                "CONFIG_TYPE": "Enum",
                "ENUM_NAME": "ScrapeTaskOption",
                "DESCRIPTION": (
                    [
                        (
                            "A scrape job is created by choosing a start date, an end date and "
                            "any data sets you wish to scrape for each day in the specified date "
                            "range. When the job is ran, a list of dates is created from the "
                            "start and end dates (both endpoints are included in the list).\n"
                        ),
                        (
                            "If SCRAPE_TASK_OPTION == BY_DATE (default behavior), the job is "
                            "executed by iterating through the list of dates. For each data set "
                            "a list of URLs to scrape for the current date is generated. After "
                            "all data sets specified for this job have been scraped, the job "
                            "moves to the next date in the list and scrapes all data sets for "
                            "that date. After all days in the list have been scraped, the job "
                            "ends.\n"
                        ),
                        (
                            "If SCRAPE_TASK_OPTION == BY_DATA_SET, the job is executed by "
                            "iterating through the list of data sets to scrape. For each data "
                            "set, a list of URLs to scrape for all days in the range specified "
                            "by the start and end dates is generated. After all URLs for the "
                            "data set have been scraped, the job moves on to the next data set "
                            "and scrapes all URLs for the specified date range. After all data "
                            "sets have been scraped, the job ends."
                        ),
                    ],
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "URL_SCRAPE_DELAY": {
                "CONFIG_TYPE": "Numeric",
                "CLASS_NAME": "UrlScrapeDelay",
                "DESCRIPTION": (
                    "As a common courtesy (a.k.a, to avoid being banned), after scraping a "
                    "webpage you should wait for a few seconds before requesting the next "
                    "URL. You can specify a single length of time to use for all URLs, or "
                    "create random delay lengths by specifying a minimum and maximum length "
                    "of time."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "BATCH_JOB_SETTINGS": {
                "CONFIG_TYPE": "Numeric",
                "CLASS_NAME": "BatchJobSettings",
                "DESCRIPTION": (
                    "Number of URLs to scrape per batch. You can specify a single amount to "
                    "use for all batches, or create random batch sizes by specifying a "
                    "minimum and maximum batch size."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "BATCH_SCRAPE_DELAY": {
                "CONFIG_TYPE": "Numeric",
                "CLASS_NAME": "BatchScrapeDelay",
                "DESCRIPTION": (
                    "Some websites will ban you even if you wait for a few seconds between "
                    "each request. To avoid being banned, you can scrape URLs in batches "
                    "(recommended batch size: ~50 URLs/batch) and wait for a long period of "
                    "time (30-45 minutes) before you begin a new batch. You can specify a "
                    "single length of time to use for all batches or create random delay "
                    "lengths by specifying a minimum and maximum length of time."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": False,
            },
            "HTML_STORAGE": {
                "CONFIG_TYPE": "Enum",
                "ENUM_NAME": "HtmlStorageOption",
                "DESCRIPTION": (
                    "By default, HTML is NOT saved after it has been parsed. However, you "
                    "can choose to save scraped HTML in a local folder, S3 bucket, or "
                    "both."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "HTML_LOCAL_FOLDER_PATH": {
                "CONFIG_TYPE": "Path",
                "CLASS_NAME": "LocalFolderPathSetting",
                "DESCRIPTION": (
                    "Local folder path where scraped HTML should be stored. The application "
                    "will always check for saved HTML content in this location before "
                    "sending a request to the website."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "HTML_S3_FOLDER_PATH": {
                "CONFIG_TYPE": "Path",
                "CLASS_NAME": "S3FolderPathSetting",
                "DESCRIPTION": (
                    "Path to a folder within an S3 bucket where scraped HTML should be "
                    "stored. The application will always check for saved HTML content in "
                    "this location before sending a request to the website."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "JSON_STORAGE": {
                "CONFIG_TYPE": "Enum",
                "ENUM_NAME": "JsonStorageOption",
                "DESCRIPTION": (
                    "MLB data is parsed from HTML and stored in JSON docs. You can store "
                    "the JSON docs in a local folder, S3 bucket, or both."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "JSON_LOCAL_FOLDER_PATH": {
                "CONFIG_TYPE": "Path",
                "CLASS_NAME": "LocalFolderPathSetting",
                "DESCRIPTION": "Local folder path where parsed JSON data should be stored.",
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "JSON_S3_FOLDER_PATH": {
                "CONFIG_TYPE": "Path",
                "CLASS_NAME": "S3FolderPathSetting",
                "DESCRIPTION": ("Path to a folder within an S3 bucket where parsed JSON data should be stored."),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "SCRAPED_DATA_COMBINE_CONDITION": {
                "CONFIG_TYPE": "Enum",
                "ENUM_NAME": "ScrapeCondition",
                "DESCRIPTION": (
                    "By default, the process to combine and audit all scraped data for a "
                    "single game is done only once (ONLY_MISSING_DATA). You can overwrite "
                    "existing data by selecting ALWAYS, or prevent any data from being combined "
                    "by selecting NEVER."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "COMBINED_DATA_STORAGE": {
                "CONFIG_TYPE": "Enum",
                "ENUM_NAME": "CombinedDataStorageOption",
                "DESCRIPTION": (
                    "When all data sets have been scraped for a single game, the parsed data is "
                    "audited and combined into a single JSON file. You can choose to store these "
                    "files on your local file system, on S3, or both."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "COMBINED_DATA_LOCAL_FOLDER_PATH": {
                "CONFIG_TYPE": "Path",
                "CLASS_NAME": "LocalFolderPathSetting",
                "DESCRIPTION": ("Local folder path where files containing combined game data should be stored."),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "COMBINED_DATA_S3_FOLDER_PATH": {
                "CONFIG_TYPE": "Path",
                "CLASS_NAME": "S3FolderPathSetting",
                "DESCRIPTION": (
                    "Path to a folder within an S3 bucket where files containing combined game "
                    "data should be stored."
                ),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
            "DB_BACKUP_FOLDER_PATH": {
                "CONFIG_TYPE": "Path",
                "CLASS_NAME": "LocalFolderPathSetting",
                "DESCRIPTION": ("Local folder path where csv files containing exported table data should be stored."),
                "SAME_SETTING_FOR_ALL_DATA_SETS": True,
            },
        }



"""Scrape brooksbaseball daily dashboard page."""
import uuid
from collections import defaultdict, OrderedDict
from string import Template

from lxml import html

from vigorish.constants import (
    PITCH_DES_DID_SWING,
    PITCH_DES_MADE_CONTACT,
    PITCH_DES_SWINGING_STRIKE,
    ZONE_LOCATIONS_INSIDE_SZ,
)
from vigorish.scrape.brooks_pitchfx.models.pitchfx import BrooksPitchFxData
from vigorish.scrape.brooks_pitchfx.models.pitchfx_log import BrooksPitchFxLog
from vigorish.util.dt_format_strings import DT_AWARE
from vigorish.util.result import Result

PITCHFX_COLUMN_NAMES = "//tr/th/text()"
PITCHFX_TABLE_ROWS = '//tr[@style="white-space:nowrap"]'
T_PITCHFX_MEASUREMENT = "td[${i}]//text()"
FILTER_NAMES = ["datestamp", "gid", "pitcher_team"]


def parse_pitchfx_log(scraped_html, pitch_log):
    page_content = html.fromstring(scraped_html, base_url=pitch_log.pitchfx_url)
    pitchfx_log_dict = {}
    result = parse_pitchfx_table(page_content, pitch_log)
    if result.failure:
        return result
    pitchfx_data = result.value
    pitchfx_log_dict["pitchfx_log"] = pitchfx_data
    pitchfx_log_dict["pitch_count_by_inning"] = get_pitch_count_by_inning(pitchfx_data)
    pitchfx_log_dict["total_pitch_count"] = len(pitchfx_data)
    pitchfx_log_dict["pitcher_name"] = pitch_log.pitcher_name
    pitchfx_log_dict["pitcher_id_mlb"] = pitch_log.pitcher_id_mlb
    pitchfx_log_dict["pitch_app_id"] = pitch_log.pitch_app_id
    pitchfx_log_dict["pitcher_team_id_bb"] = pitch_log.pitcher_team_id_bb
    pitchfx_log_dict["opponent_team_id_bb"] = pitch_log.opponent_team_id_bb
    pitchfx_log_dict["bb_game_id"] = pitch_log.bb_game_id
    pitchfx_log_dict["bbref_game_id"] = pitch_log.bbref_game_id
    pitchfx_log_dict["game_date_year"] = pitch_log.game_date_year
    pitchfx_log_dict["game_date_month"] = pitch_log.game_date_month
    pitchfx_log_dict["game_date_day"] = pitch_log.game_date_day
    pitchfx_log_dict["game_time_hour"] = pitch_log.game_time_hour
    pitchfx_log_dict["game_time_minute"] = pitch_log.game_time_minute
    pitchfx_log_dict["time_zone_name"] = pitch_log.time_zone_name
    pitchfx_log_dict["pitchfx_url"] = pitch_log.pitchfx_url
    pitchfx_log = BrooksPitchFxLog(**pitchfx_log_dict)
    return Result.Ok(pitchfx_log)


def get_pitch_count_by_inning(pfx_data):
    unordered = defaultdict(int)
    for pfx in pfx_data:
        unordered[pfx.inning] += 1
    pitch_count_by_inning = OrderedDict()
    for k in sorted(unordered.keys()):
        pitch_count_by_inning[k] = unordered[k]
    return pitch_count_by_inning


def parse_pitchfx_table(page_content, pitch_log):
    pitchfx_data = []
    column_names = page_content.xpath(PITCHFX_COLUMN_NAMES)
    table_rows = page_content.xpath(PITCHFX_TABLE_ROWS)
    for i, row in enumerate(table_rows):
        result = parse_pitchfx_data(column_names, row, i, pitch_log)
        if result.failure:
            return result
        pitchfx = result.value
        pitchfx_data.append(pitchfx)
    return fix_missing_des(pitchfx_data, pitch_log)


def parse_pitchfx_data(column_names, table_row, row_num, pitch_log):
    pitchfx_dict = {}
    for i, name in enumerate(column_names):
        if name.lower() in FILTER_NAMES:
            continue
        query = Template(T_PITCHFX_MEASUREMENT).substitute(i=(i + 1))
        results = table_row.xpath(query)
        if not results:
            if name == "des":
                pitchfx_dict["des"] = "missing_des"
                continue
            elif name == "pdes":
                pitchfx_dict["pdes"] = "missing_pdes"
                continue
            elif name == "play_guid":
                pitchfx_dict["play_guid"] = str(uuid.uuid4())
                continue
            elif name == "zone_location":
                pitchfx_dict["zone_location"] = 99
                continue
            error = (
                f"Error occurred attempting to parse '{name}' (column #{i}) from pitchfx table:\n"
                f"Game ID.......: {pitch_log.bbref_game_id}\n"
                f"Pitcher.......: {pitch_log.pitcher_name} ({pitch_log.pitcher_id_mlb})\n"
                f"Row Number....: {row_num}\n"
                f"PitchFX URL...: {pitch_log.pitchfx_url}\n"
                f"Partial Data..: {pitchfx_dict}"
            )
            return Result.Fail(error)
        pitchfx_dict[name.lower()] = convert_type_for_dataclass(name.lower(), results[0])
    pitchfx_dict["pitcher_name"] = pitch_log.pitcher_name
    pitchfx_dict["pitch_app_id"] = pitch_log.pitch_app_id
    pitchfx_dict["pitcher_team_id_bb"] = pitch_log.pitcher_team_id_bb
    pitchfx_dict["opponent_team_id_bb"] = pitch_log.opponent_team_id_bb
    pitchfx_dict["bb_game_id"] = pitch_log.bb_game_id
    pitchfx_dict["bbref_game_id"] = pitch_log.bbref_game_id
    pitchfx_dict["table_row_number"] = row_num
    pitchfx = BrooksPitchFxData(**pitchfx_dict)

    game_start_time = pitch_log.game_start_time
    pitchfx.game_start_time_str = game_start_time.strftime(DT_AWARE) if game_start_time else ""
    pitchfx.has_zone_location = pitchfx.zone_location != 99
    pitchfx.batter_did_swing = pitchfx_dict["pdes"] in PITCH_DES_DID_SWING
    pitchfx.batter_made_contact = pitchfx_dict["pdes"] in PITCH_DES_MADE_CONTACT
    pitchfx.called_strike = "Called Strike" in pitchfx_dict["pdes"]
    pitchfx.swinging_strike = pitchfx_dict["pdes"] in PITCH_DES_SWINGING_STRIKE
    if pitchfx.has_zone_location:
        pitchfx.inside_strike_zone = pitchfx.zone_location in ZONE_LOCATIONS_INSIDE_SZ
        pitchfx.outside_strike_zone = not pitchfx.inside_strike_zone
        pitchfx.swing_inside_zone = pitchfx.batter_did_swing and pitchfx.inside_strike_zone
        pitchfx.swing_outside_zone = pitchfx.batter_did_swing and pitchfx.outside_strike_zone
        pitchfx.contact_inside_zone = pitchfx.batter_made_contact and pitchfx.inside_strike_zone
        pitchfx.contact_outside_zone = pitchfx.batter_made_contact and pitchfx.outside_strike_zone
    return Result.Ok(pitchfx)


def convert_type_for_dataclass(name, value):
    dataclass_field = BrooksPitchFxData.__dataclass_fields__.get(name)
    field_type = dataclass_field.type if dataclass_field else str
    return (
        True
        if field_type is bool and value
        else False
        if field_type is bool and not value
        else None
        if not value
        else int(value)
        if field_type is int and not isinstance(value, int)
        else float(value)
        if field_type is float and not isinstance(value, float)
        else value
    )


def fix_missing_des(pfx_data, pitch_log):
    missing_des = any(pfx.des == "missing_des" for pfx in pfx_data)
    if not missing_des:
        return Result.Ok(pfx_data)
    fix_ab_ids = {pfx.ab_id for pfx in pfx_data if pfx.des == "missing_des"}
    for ab_id in list(fix_ab_ids):
        missing_des = [pfx for pfx in pfx_data if this_ab_and_missing_des(pfx, ab_id)]
        valid_des = list({pfx.des for pfx in pfx_data if this_ab_and_has_des(pfx, ab_id)})
        if not valid_des:
            continue
        if len(valid_des) == 1:
            for pfx in missing_des:
                pfx.des = valid_des[0]
            continue
        error = (
            f"Unable to fix missing description for pitchfx ab_id={ab_id}\n"
            f"Game ID.......: {pitch_log.bbref_game_id}\n"
            f"Pitcher.......: {pitch_log.pitcher_name} ({pitch_log.pitcher_id_mlb})\n"
            f"PitchFX URL...: {pitch_log.pitchfx_url}\n"
            f"des values....: {valid_des}"
        )
        return Result.Fail(error)
    return Result.Ok(pfx_data)


def this_ab_and_missing_des(pfx, ab_id):
    return pfx.ab_id == ab_id and pfx.des == "missing_des"


def this_ab_and_has_des(pfx, ab_id):
    return pfx.ab_id == ab_id and pfx.des != "missing_des"



import json

from vigorish.enums import DataSet
from vigorish.scrape.url_builder import create_url_set as get_url_set
from vigorish.util.list_helpers import flatten_list2d
from vigorish.util.result import Result


class UrlTracker:
    def __init__(self, db_job, data_set, scraped_data):
        self.db_job = db_job
        self.data_set = data_set
        self.scraped_data = scraped_data
        self.all_urls = []
        self.need_urls = []
        self.missing_urls = []
        self.cached_urls = []
        self.completed_urls = []
        self.skip_urls = []
        self.missing_urls_filepath = self.db_job.url_set_filepath
        self.scraped_html_folderpath = self.db_job.scraped_html_folders[self.data_set]

    @property
    def total_urls(self):
        return len(flatten_list2d(list(self.all_urls.values()))) if self.all_urls else 0

    @property
    def parse_urls(self):
        return self.missing_urls + self.completed_urls + self.cached_urls

    @property
    def parse_url_ids(self):
        return (
            [url.url_id for url in self.parse_urls]
            if self.data_set != DataSet.BROOKS_PITCH_LOGS
            else list({url.url_id[:12] for url in self.parse_urls})
        )

    @property
    def html_scraping_complete(self):
        return all(url.html_was_scraped for url in self.parse_urls)

    @property
    def identify_html_report(self):
        percent_complete = 0
        total_remaining = 0
        total_complete = len(self.skip_urls) + len(self.need_urls)
        if self.all_urls:
            percent_complete = total_complete / float(self.total_urls)
            total_remaining = self.total_urls - total_complete
        return (
            f"Determining URLs to scrape... {percent_complete:.0%} "
            f"({len(self.skip_urls)} Skip, {len(self.need_urls)} Scrape, "
            f"{total_remaining} Remaining)"
        )

    @property
    def retrieve_html_report(self):
        total_complete = len(self.skip_urls) + len(self.cached_urls) + len(self.missing_urls)
        percent_complete = total_complete / float(self.total_urls)
        return (
            f"Retrieving scraped HTML... {percent_complete:.0%} "
            f"({len(self.skip_urls)} Skipped, {len(self.cached_urls)} Found, "
            f"{len(self.missing_urls)} Missing, {len(self.need_urls)} Remaining)"
        )

    @property
    def scrape_html_report(self):
        report = f"Scraping missing HTML... ({len(self.skip_urls)} Skipped,"
        if self.completed_urls:
            report = f"{report} {len(self.completed_urls)} Scraped,"
        return f"{report} {len(self.cached_urls)} Found, {len(self.missing_urls)} Missing)"

    @property
    def save_html_report(self):
        total_urls = len(self.missing_urls) + len(self.completed_urls)
        percent_complete = len(self.completed_urls) / float(total_urls)
        return f"Saving scraped HTML... {percent_complete:.0%} ({len(self.completed_urls)}/{total_urls}) URLs"

    def create_url_set(self, start_date, end_date):
        result = get_url_set(start_date, end_date, self.db_job, self.data_set, self.scraped_data)
        if result.failure:
            return result
        self.all_urls = result.value
        return Result.Ok()

    def parse_html_report(self, parsed_count, game_id=None):
        percent_complete = 0
        if len(self.parse_url_ids) > 0:
            percent_complete = parsed_count / float(len(self.parse_url_ids))
        if game_id:
            data_set_str = "PitchFX Logs"
            unit = "URLs"
            if self.data_set == DataSet.BROOKS_PITCH_LOGS:
                data_set_str = "Pitch Logs"
                unit = "Games"
            return (
                f"Parsing {data_set_str} for {game_id}... {percent_complete:.0%} "
                f"({parsed_count}/{len(self.parse_url_ids)}) {unit}"
            )
        return f"Parsing scraped HTML... {percent_complete:.0%} ({parsed_count}/{len(self.parse_url_ids)}) URLs"

    def create_missing_urls_json_file(self):
        urls_dict = [url.as_dict() for url in self.missing_urls]
        urls_json = json.dumps(urls_dict, indent=2, sort_keys=False)
        self.missing_urls_filepath.write_text(urls_json)
        return self.missing_urls_filepath

    def get_html(self, url_id):
        match = [url for url in self.parse_urls if url.url_id == url_id]
        return match[0].html if match else None

    def remove_scraped_html(self):
        for url in self.completed_urls:
            if url.scraped_file_path.exists():
                url.scraped_file_path.unlink()



import vigorish.database as db
from vigorish.enums import DataSet, ScrapeCondition
from vigorish.scrape.brooks_pitch_logs.models.pitch_logs_for_game import BrooksPitchLogsForGame
from vigorish.scrape.brooks_pitch_logs.parse_html import parse_pitch_log
from vigorish.scrape.scrape_task import ScrapeTaskABC
from vigorish.status.update_status_brooks_pitch_logs import update_status_brooks_pitch_logs_for_game
from vigorish.util.dt_format_strings import DATE_MONTH_NAME, DATE_ONLY_2
from vigorish.util.result import Result


class ScrapeBrooksPitchLogs(ScrapeTaskABC):
    def __init__(self, app, db_job):
        self.data_set = DataSet.BROOKS_PITCH_LOGS
        self.req_data_set = DataSet.BROOKS_GAMES_FOR_DATE
        super().__init__(app, db_job)

    def check_prerequisites(self, game_date):
        brooks_games_for_date = db.DateScrapeStatus.verify_brooks_daily_dashboard_scraped_for_date(
            self.db_session, game_date
        )
        if brooks_games_for_date:
            return Result.Ok()
        date_str = game_date.strftime(DATE_ONLY_2)
        error = (
            f"Brooks games for date {date_str} have not been scraped, unable to scrape Brooks "
            "pitch logs until this has been done."
        )
        return Result.Fail(error)

    def check_current_status(self, game_date):
        if self.scrape_condition == ScrapeCondition.ALWAYS:
            return Result.Ok()
        scraped_brooks_pitch_logs = db.DateScrapeStatus.verify_all_brooks_pitch_logs_scraped_for_date(
            self.db_session, game_date
        )
        return Result.Ok() if not scraped_brooks_pitch_logs else Result.Fail("skip")

    def parse_scraped_html(self):
        parsed = 0
        for game_date in self.date_range:
            brooks_games_for_date = self.scraped_data.get_brooks_games_for_date(game_date)
            if not brooks_games_for_date:
                date_str = game_date.strftime(DATE_MONTH_NAME)
                return Result.Fail(f"Failed to retrieve {self.req_data_set} (Date: {date_str})")
            for game in brooks_games_for_date.games:
                if game.bbref_game_id not in self.url_tracker.parse_url_ids:
                    continue
                if game.might_be_postponed:
                    continue
                self.spinner.text = self.url_tracker.parse_html_report(parsed, game.bbref_game_id)
                pitch_logs_for_game = BrooksPitchLogsForGame()
                pitch_logs_for_game.bb_game_id = game.bb_game_id
                pitch_logs_for_game.bbref_game_id = game.bbref_game_id
                pitch_logs_for_game.pitch_log_count = game.pitcher_appearance_count
                scraped_pitch_logs = []
                for pitcher_id, url in game.pitcher_appearance_dict.items():
                    pitch_app_id = f"{game.bbref_game_id}_{pitcher_id}"
                    html = self.url_tracker.get_html(pitch_app_id)
                    result = parse_pitch_log(html, game, pitcher_id, url)
                    if result.failure:
                        return result
                    pitch_log = result.value
                    scraped_pitch_logs.append(pitch_log)
                pitch_logs_for_game.pitch_logs = scraped_pitch_logs
                result = self.scraped_data.save_json(self.data_set, pitch_logs_for_game)
                if result.failure:
                    return result
                result = self.update_status(pitch_logs_for_game)
                if result.failure:
                    return result
                parsed += 1
                self.spinner.text = self.url_tracker.parse_html_report(parsed, game.bbref_game_id)
                self.db_session.commit()
        return Result.Ok()

    def parse_html(self, url_details):
        pass

    def update_status(self, parsed_data):
        return update_status_brooks_pitch_logs_for_game(self.db_session, parsed_data)



from sqlalchemy import func, join, select
from sqlalchemy_utils import create_view

import vigorish.database as db


class Game_PitchApp_View(db.Base):
    __table__ = create_view(
        name="game_pitch_app_status",
        selectable=select(
            [
                db.GameScrapeStatus.id.label("id"),
                db.GameScrapeStatus.bbref_game_id.label("bbref_game_id"),
                func.count(db.PitchAppScrapeStatus.id).label("total_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.scraped_pitchfx).label("total_pitchfx_scraped"),
                func.sum(db.PitchAppScrapeStatus.no_pitchfx_data).label("total_no_pitchfx_data"),
                func.sum(db.PitchAppScrapeStatus.combined_pitchfx_bbref_data).label(
                    "total_combined_pitchfx_bbref_data"
                ),
                func.sum(db.PitchAppScrapeStatus.batters_faced_bbref).label("total_batters_faced_bbref"),
                func.sum(db.PitchAppScrapeStatus.batters_faced_pitchfx).label("total_batters_faced_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.pitch_count_pitch_log).label("total_pitch_count_pitch_log"),
                func.sum(db.PitchAppScrapeStatus.pitch_count_bbref).label("total_pitch_count_bbref"),
                func.sum(db.PitchAppScrapeStatus.pitch_count_pitchfx).label("total_pitch_count_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.pitch_count_pitchfx_audited).label(
                    "total_pitch_count_pitchfx_audited"
                ),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_pitchfx_complete).label(
                    "total_at_bats_pitchfx_complete"
                ),
                func.sum(db.PitchAppScrapeStatus.patched_pitchfx_count).label("total_patched_pitchfx_count"),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_patched_pitchfx).label("total_at_bats_patched_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.missing_pitchfx_count).label("total_missing_pitchfx_count"),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_missing_pitchfx).label("total_at_bats_missing_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.removed_pitchfx_count).label("total_removed_pitchfx_count"),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_removed_pitchfx).label("total_at_bats_removed_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.pitchfx_error).label("total_pitchfx_error"),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_pitchfx_error).label("total_at_bats_pitchfx_error"),
                func.sum(db.PitchAppScrapeStatus.invalid_pitchfx).label("total_invalid_pitchfx"),
                func.sum(db.PitchAppScrapeStatus.invalid_pitchfx_count).label("total_invalid_pitchfx_count"),
                func.sum(db.PitchAppScrapeStatus.total_at_bats_invalid_pitchfx).label("total_at_bats_invalid_pitchfx"),
            ]
        )
        .select_from(
            join(
                db.GameScrapeStatus,
                db.PitchAppScrapeStatus,
                db.GameScrapeStatus.id == db.PitchAppScrapeStatus.scrape_status_game_id,
            )
        )
        .group_by(db.GameScrapeStatus.id),
        metadata=db.Base.metadata,
        cascade_on_drop=False,
    )



"""Menu that allows the user to view all jobs grouped by status."""
import vigorish.database as db
from vigorish.cli.menu import Menu
from vigorish.cli.menu_items.return_to_parent import ReturnToParent
from vigorish.cli.menus.jobs_menu import JobsMenu
from vigorish.constants import EMOJIS


class AllJobsMenu(Menu):
    def __init__(self, app):
        super().__init__(app)
        self.menu_text = "Jobs are grouped according to their current status:"
        self.menu_item_text = "View All Jobs"
        self.menu_item_emoji = EMOJIS.get("ROBOT", "")

    def populate_menu_items(self):
        jobs_grouped = db.ScrapeJob.get_all_jobs_grouped_sorted(self.db_session)
        self.menu_items = [
            JobsMenu(self.app, jobs_grouped[status], status, menu_number)
            for menu_number, status in enumerate(jobs_grouped.keys(), start=1)
        ]
        self.menu_items.append(ReturnToParent(self.app, "Return to Main Menu"))



# flake8: noqa
from vigorish.cli.menus.admin_tasks_menu import AdminTasksMenu
from vigorish.cli.menus.all_jobs_menu import AllJobsMenu
from vigorish.cli.menus.config_settings_menu import ConfigSettingsMenu
from vigorish.cli.menus.env_var_settings_menu import EnvVarSettingsMenu
from vigorish.cli.menus.jobs_menu import JobsMenu
from vigorish.cli.menus.scraped_data_errors_menu import ScrapedDataErrorsMenu
from vigorish.cli.menus.settings_menu import SettingsMenu
from vigorish.cli.menus.view_game_data_menu import ViewGameDataMenu



"""View all scraped data for a single game."""
import subprocess

from halo import Halo

from vigorish.cli.components.prompts import user_options_prompt
from vigorish.cli.components.util import (
    get_random_cli_color,
    get_random_dots_spinner,
    print_heading,
    print_message,
)
from vigorish.cli.menu_item import MenuItem
from vigorish.constants import EMOJIS, MENU_NUMBERS
from vigorish.data.game_data import GameData
from vigorish.util.result import Result


class ViewGameData(MenuItem):
    def __init__(self, app, bbref_game_id):
        super().__init__(app)
        self.bbref_game_id = bbref_game_id
        self.menu_item_text = bbref_game_id
        self.menu_item_emoji = EMOJIS.get("NEWSPAPER")

    @property
    def away_team_id(self):
        return self.game_data.away_team_id

    @property
    def home_team_id(self):
        return self.game_data.home_team_id

    def launch(self):
        subprocess.run(["clear"])
        self.load_boxscore_data()
        while True:
            self.print_matchup_and_linescore()
            result = self.select_data_prompt()
            if result.failure:
                break
            (team_id, menu_option) = result.value
            if menu_option == "META_INFO":
                table_viewer = self.game_data.view_game_meta_info()
                table_viewer.launch()
                continue
            if menu_option == "ALL":
                innings_viewer = self.game_data.view_at_bats_by_inning().value
                result = self.view_at_bats_by_inning(innings_viewer)
                if result.failure:
                    break
                continue
            result = self.view_boxscore(team_id, menu_option)
            if result.failure:
                break
        return Result.Ok()

    def load_boxscore_data(self):
        spinner = Halo(spinner=get_random_dots_spinner(), color=get_random_cli_color())
        spinner.text = f"Loading data for {self.bbref_game_id}..."
        spinner.start()
        self.game_data = GameData(self.app, self.bbref_game_id)
        self.game_data.bat_boxscore[self.away_team_id]
        self.game_data.pitch_boxscore[self.away_team_id]
        self.game_data.bat_boxscore[self.home_team_id]
        self.game_data.pitch_boxscore[self.home_team_id]
        spinner.stop()

    def print_matchup_and_linescore(self):
        subprocess.run(["clear"])
        matchup = self.game_data.get_matchup_details()
        linescore = self.game_data.get_tui_linescore()
        print_heading(f"Scraped Data Viewer for Game ID: {self.bbref_game_id}", fg="bright_yellow")
        print_message(matchup, fg="bright_cyan", bold=True, wrap=False)
        print_message(linescore, fg="bright_cyan", wrap=False)

    def select_data_prompt(self):
        prompt = (
            f"\nData for {self.bbref_game_id} can be viewed in several different ways, "
            "please choose an option from the list below:"
        )
        choices = {
            f"{MENU_NUMBERS.get(1)}  {self.bat_stats_text(False)}": (self.away_team_id, "BAT"),
            f"{MENU_NUMBERS.get(2)}  {self.pitch_stats_text(False)}": (self.away_team_id, "PITCH"),
            f"{MENU_NUMBERS.get(3)}  {self.bat_stats_text(True)}": (self.home_team_id, "BAT"),
            f"{MENU_NUMBERS.get(4)}  {self.pitch_stats_text(True)}": (self.home_team_id, "PITCH"),
            f"{MENU_NUMBERS.get(5)}  All At Bats By Inning": (None, "ALL"),
            f"{MENU_NUMBERS.get(6)}  Game Meta Information": (None, "META_INFO"),
            f"{EMOJIS.get('BACK')} Return to Previous Menu": None,
        }
        return user_options_prompt(choices, prompt, clear_screen=False)

    def pitch_stats_text(self, is_home_team):
        team_id = self.home_team_id if is_home_team else self.away_team_id
        return f"{team_id} Pitcher Boxscore"

    def bat_stats_text(self, is_home_team):
        team_id = self.home_team_id if is_home_team else self.away_team_id
        return f"{team_id} Batter Boxscore"

    def view_at_bats_by_inning(self, innings_viewer):
        subprocess.run(["clear"])
        while True:
            result = self.innings_viewer_prompt(innings_viewer)
            if result.failure:
                break
            table_viewer = result.value
            table_viewer.launch()
        return Result.Ok()

    def innings_viewer_prompt(self, innings_viewer):
        choices = {f"{EMOJIS.get('ASTERISK')}  All Innings": innings_viewer["ALL"]}
        for inning, table_viewer in innings_viewer.items():
            if inning == "ALL":
                continue
            choices[f"{MENU_NUMBERS.get(int(inning[-2:]), inning[-2:])}  {inning}"] = table_viewer
        choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
        prompt = "Select an inning to view:"
        return user_options_prompt(choices, prompt)

    def view_boxscore(self, team_id, player_type):
        boxscore = self.get_boxscore(team_id, player_type)
        while True:
            subprocess.run(["clear"])
            result = self.select_player_prompt(player_type, boxscore)
            if result.failure:
                return Result.Ok()
            mlb_id = result.value
            if player_type == "BAT":
                self.view_at_bats_for_player(player_type, mlb_id)
            else:
                subprocess.run(["clear"])
                result = self.select_pitcher_data_prompt(mlb_id)
                if result.failure:
                    return Result.Ok()
                pitcher_data = result.value
                if pitcher_data == "AT_BATS":
                    self.view_at_bats_for_player(player_type, mlb_id)
                if pitcher_data == "BAT_STATS":
                    self.view_bat_stats_by_pitch_type_for_player(mlb_id)
                if pitcher_data == "PITCH_MIX_BY_STANCE":
                    self.view_pitch_mix_batter_stance_splits(mlb_id)
                if pitcher_data == "PITCH_MIX_BY_SEASON":
                    self.view_pitch_mix_season_splits(mlb_id)
                if pitcher_data == "PLATE_DISCIPLINE":
                    self.view_pd_pitch_type_splits_for_pitcher(mlb_id)
                if pitcher_data == "BATTED_BALL":
                    self.view_bb_pitch_type_splits_for_pitcher(mlb_id)

    def get_boxscore(self, team_id, player_type):
        boxscore_dict = {
            "PITCH": self.game_data.pitch_boxscore,
            "BAT": self.game_data.bat_boxscore,
        }
        return boxscore_dict[player_type][team_id]

    def select_player_prompt(self, player_type, boxscore):
        player_prompt_dict = {
            "PITCH": self.select_pitcher_prompt,
            "BAT": self.select_batter_prompt,
        }
        return player_prompt_dict[player_type](boxscore)

    def view_at_bats_for_player(self, player_type, mlb_id):
        view_at_bats_dict = {
            "PITCH": self.view_at_bats_for_pitcher,
            "BAT": self.view_at_bats_for_batter,
        }
        return view_at_bats_dict[player_type](mlb_id)

    def select_pitcher_prompt(self, pitch_boxscore):
        max_name_length = self.get_name_max_length(pitch_boxscore)
        choices = {
            f"{self.select_pitcher_text(num, box, max_name_length)}": box["mlb_id"]
            for num, box in enumerate(pitch_boxscore.values(), start=1)
        }
        choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
        prompt = (
            "All pitchers that apppeared in the game are listed below. Select a player and "
            "press ENTER to access information for this start and various career stats:"
        )
        print_heading(f"Scraped Data Viewer for Game ID: {self.bbref_game_id}", fg="bright_yellow")
        return user_options_prompt(choices, prompt, auto_scroll=False, clear_screen=False)

    def get_name_max_length(self, boxscore):
        return max(len(box["name"]) for box in boxscore.values())

    def select_pitcher_text(self, num, box, max_name_length):
        fill1_count = 2 if num < 10 else 1
        fill2_count = max_name_length - len(box["name"])
        return (
            f"{MENU_NUMBERS.get(num, num)}{' '*fill1_count}[{box['pitch_app_type']}] "
            f"{box['name']}{'.'*fill2_count}..: "
            f"{box['game_results']}"
        )

    def select_pitcher_data_prompt(self, mlb_id):
        pitcher_name = self.game_data.get_player_id_map(mlb_id=mlb_id).mlb_name
        prompt = (
            f"Data for {pitcher_name} can be viewed in several different ways, "
            "please choose an option from the list below:"
        )
        choices = {
            f"{MENU_NUMBERS.get(1)}  Play-by-Play Data for All At Bats": "AT_BATS",
            f"{MENU_NUMBERS.get(2)}  Batting Stats By Pitch Type": "BAT_STATS",
            f"{MENU_NUMBERS.get(3)}  Pitch Mix Split by Batter Stance": "PITCH_MIX_BY_STANCE",
            f"{MENU_NUMBERS.get(4)}  Pitch Mix Split by Season": "PITCH_MIX_BY_SEASON",
            f"{MENU_NUMBERS.get(5)}  Plate Discipline Metrics Split by Pitch Type": "PLATE_DISCIPLINE",
            f"{MENU_NUMBERS.get(6)}  Batted Ball Metrics Split by Pitch Type": "BATTED_BALL",
            f"{EMOJIS.get('BACK')} Return to Previous Menu": None,
        }
        print_heading(f"Scraped Data Viewer for Game ID: {self.bbref_game_id}", fg="bright_yellow")
        return user_options_prompt(choices, prompt, clear_screen=False)

    def view_at_bats_for_pitcher(self, mlb_id):
        subprocess.run(["clear"])
        innings_viewer = self.game_data.view_valid_at_bats_for_pitcher(mlb_id).value
        return self.view_at_bats_by_inning(innings_viewer)

    def view_bat_stats_by_pitch_type_for_player(self, mlb_id):
        subprocess.run(["clear"])
        table_viewer = self.game_data.view_bat_stats_pitch_type_splits_for_pitcher(mlb_id).value
        table_viewer.launch()
        return Result.Ok()

    def view_pitch_mix_batter_stance_splits(self, mlb_id):
        subprocess.run(["clear"])
        table_viewer = self.game_data.view_pitch_mix_batter_stance_splits(mlb_id).value
        table_viewer.launch()
        return Result.Ok()

    def view_pitch_mix_season_splits(self, mlb_id):
        subprocess.run(["clear"])
        table_viewer = self.game_data.view_pitch_mix_season_splits(mlb_id).value
        table_viewer.launch()
        return Result.Ok()

    def view_pd_pitch_type_splits_for_pitcher(self, mlb_id):
        subprocess.run(["clear"])
        table_viewer = self.game_data.view_pd_pitch_type_splits_for_pitcher(mlb_id).value
        table_viewer.launch()
        return Result.Ok()

    def view_bb_pitch_type_splits_for_pitcher(self, mlb_id):
        subprocess.run(["clear"])
        table_viewer = self.game_data.view_bb_pitch_type_splits_for_pitcher(mlb_id).value
        table_viewer.launch()
        return Result.Ok()

    def select_batter_prompt(self, bat_boxscore):
        max_name_length = self.get_name_max_length(bat_boxscore)
        choices = {
            f"{self.select_batter_text(order_num, box, max_name_length)}": box["mlb_id"]
            for order_num, box in bat_boxscore.items()
        }
        choices[f"{EMOJIS.get('BACK')} Return to Previous Menu"] = None
        prompt = (
            "Starting lineup and all substitutes who made at least one plate appearance are "
            "listed below. Select a player and press ENTER to view details of each at bat:"
        )
        print_heading(f"Scraped Data Viewer for Game ID: {self.bbref_game_id}", fg="bright_yellow")
        return user_options_prompt(choices, prompt, auto_scroll=False, clear_screen=False)

    def select_batter_text(self, lineup_slot, box, max_name_len):
        fill_count = max_name_len - len(box["name"])
        def_pos_fill = "" if len(str(box["def_position"])) >= 2 else " "
        return (
            f"{self.get_emoji_for_lineup_slot(lineup_slot)} "
            f"[{str(box['def_position'])}]{def_pos_fill} "
            f"{box['name']}{'.'*fill_count}..: "
            f"{box['at_bats']} {box['bat_stats']}"
        )

    def get_emoji_for_lineup_slot(self, lineup_slot):
        emoji_num = MENU_NUMBERS.get(lineup_slot)
        if emoji_num:
            return f"{emoji_num} "
        return EMOJIS.get("CAP")

    def view_at_bats_for_batter(self, mlb_id):
        subprocess.run(["clear"])
        table_viewer = self.game_data.view_valid_at_bats_for_batter(mlb_id).value
        table_viewer.launch()
        return Result.Ok()



"""The main menu for the CLI."""
import subprocess

from halo import Halo
from pyfiglet import Figlet
from tabulate import tabulate

from vigorish import __version__
from vigorish.cli.components import (
    get_random_cli_color,
    get_random_dots_spinner,
    get_random_figlet_font,
    print_heading,
    print_message,
)
from vigorish.cli.menu import Menu
from vigorish.cli.menu_items import (
    CombineScrapedData,
    CreateJob,
    ExitProgram,
    SetupDatabase,
    StatusReport,
)
from vigorish.cli.menu_items.admin_tasks import NpmInstallUpdate, RestoreDatabase
from vigorish.cli.menus import (
    AdminTasksMenu,
    AllJobsMenu,
    ScrapedDataErrorsMenu,
    SettingsMenu,
    ViewGameDataMenu,
)
from vigorish.util.result import Result
from vigorish.util.sys_helpers import node_is_installed, node_modules_folder_exists


class MainMenu(Menu):
    def __init__(self, app):
        super().__init__(app)
        self.initialized = False
        self.audit_report = None

    @property
    def db_setup_complete(self):
        return self.app.db_setup_complete

    @property
    def initial_setup_complete(self):
        return node_is_installed() and node_modules_folder_exists() and self.db_setup_complete

    @property
    def needs_refresh(self):
        check_refresh_dict = {
            SetupDatabase: True,
            CreateJob: True,
            AllJobsMenu: True,
            CombineScrapedData: True,
            ScrapedDataErrorsMenu: True,
            AdminTasksMenu: True,
        }
        selected_menu_item = type(self.selected_menu_item)
        return check_refresh_dict.get(selected_menu_item, False)

    def launch(self):
        exit_menu = False
        subprocess.run(["clear"])
        self.check_app_status()
        self.initialized = True
        while not exit_menu:
            subprocess.run(["clear"])
            if self.needs_refresh:
                self.check_app_status()
            self.populate_menu_items()
            self.display_menu_text()
            result = self.prompt_user_for_menu_selection()
            if result.failure:
                return result
            exit_menu = result.value
        return Result.Ok(exit_menu)

    def check_app_status(self):
        if not self.db_setup_complete:
            return
        color = get_random_cli_color()
        if not self.initialized:
            f = Figlet(font=get_random_figlet_font(), width=120)
            print_message(f.renderText("vigorish"), wrap=False, fg=f"bright_{color}")
        spinner = Halo(spinner=get_random_dots_spinner(), color=color)
        spinner.text = "Updating metrics..." if self.initialized else "Loading..."
        spinner.start()
        if self.initialized:
            del self.app.audit_report
        self.audit_report = self.app.audit_report
        spinner.stop()

    def display_menu_text(self):
        print_heading(f"vigorish v{__version__}", fg="bright_yellow")
        if self.audit_report:
            self.display_audit_report()
        elif self.initial_setup_complete:
            print_message("All prerequisites are installed and database is initialized.")
        else:
            self.display_initial_task_status()
        print_message("\nMain Menu:\n", fg="bright_green", bold=True, underline=True)

    def display_audit_report(self):
        table_rows = [
            {
                "season": f"MLB {year} ({report['total_games']} games)",
                "scraped": len(report["scraped"]),
                "combined": len(report["successful"]),
                "failed": len(report["failed"]) + len(report["pfx_error"]) + len(report["invalid_pfx"]),
            }
            for year, report in self.audit_report.items()
        ]
        print_message(tabulate(table_rows, headers="keys"), wrap=False)

    def display_initial_task_status(self):
        if node_is_installed():
            print_message("Node.js Installed.............: YES", fg="bright_green", bold=True)
        else:
            print_message("Node.js Installed.............: NO", fg="bright_red", bold=True)
        if node_modules_folder_exists():
            print_message("Electron/Nightmare Installed..: YES", fg="bright_green", bold=True)
        else:
            print_message("Electron/Nightmare Installed..: NO", fg="bright_red", bold=True)
        if self.db_setup_complete:
            print_message("SQLite DB Initialized.........: YES", fg="bright_green", bold=True)
        else:
            print_message("SQLite DB Initialized.........: NO", fg="bright_red", bold=True)

    def populate_menu_items(self):
        main_menu_items = self.get_menu_items()
        self.menu_items = list(main_menu_items.values())
        if node_modules_folder_exists():
            self.menu_items.remove(main_menu_items["npm_install"])
        if not self.db_setup_complete:
            self.menu_items.remove(main_menu_items["create_job"])
            self.menu_items.remove(main_menu_items["all_jobs"])
            self.menu_items.remove(main_menu_items["status_reports"])
        else:
            self.menu_items.remove(main_menu_items["setup_db"])
            self.menu_items.remove(main_menu_items["restore_db"])
        if not self.audit_report:
            self.menu_items.remove(main_menu_items["combine_data"])
        if self.audit_report and not self.data_failures_exist():
            self.menu_items.remove(main_menu_items["scraped_data_errors"])
        if self.audit_report and not self.games_combined_exist():
            self.menu_items.remove(main_menu_items["scraped_game_data"])

    def prompt_user_for_menu_selection(self):
        menu = self._get_bullet_menu()
        menu.pos = self.selected_menu_item_pos
        self.selected_menu_item_text = menu.launch()
        result = self.selected_menu_item.launch()
        exit_menu = self.selected_menu_item.exit_menu
        return Result.Ok(exit_menu) if result.success else result

    def get_menu_items(self):
        return {
            "npm_install": NpmInstallUpdate(self.app),
            "setup_db": SetupDatabase(self.app),
            "restore_db": RestoreDatabase(self.app),
            "create_job": CreateJob(self.app),
            "all_jobs": AllJobsMenu(self.app),
            "combine_data": CombineScrapedData(self.app),
            "scraped_data_errors": ScrapedDataErrorsMenu(self.app),
            "scraped_game_data": ViewGameDataMenu(self.app),
            "status_reports": StatusReport(self.app),
            "settings": SettingsMenu(self.app),
            "admin_tasks": AdminTasksMenu(self.app),
            "exit_program": ExitProgram(self.app),
        }

    def data_failures_exist(self):
        audit_failures_exist = any(audit["failed"] for audit in self.audit_report.values())
        pfx_errors_exist = any(audit["pfx_error"] for audit in self.audit_report.values())
        invalid_pfx_errors_exist = any(audit["invalid_pfx"] for audit in self.audit_report.values())
        return audit_failures_exist or pfx_errors_exist or invalid_pfx_errors_exist

    def games_combined_exist(self):
        return any(audit["successful"] for audit in self.audit_report.values())



"""Helpful datetime converters and formatters."""
import time
from datetime import date, datetime, timedelta, timezone

from dateutil import tz

from vigorish.util.dt_format_strings import DATE_ONLY_2, DT_AWARE, DT_NAIVE

TIME_ZONE_LA = tz.gettz("America/Los_Angeles")
TIME_ZONE_NEW_YORK = tz.gettz("America/New_York")
TIME_SPAN_ONE_DAY = timedelta(days=1)


def get_date_range(start, end, inc=TIME_SPAN_ONE_DAY):
    if start > end:
        start_str = start.strftime(DATE_ONLY_2)
        end_str = end.strftime(DATE_ONLY_2)
        raise ValueError(f"Start date ({start_str}) must be earlier than end date ({end_str})")

    result = []
    current = start
    while current <= end:
        result.append(current)
        current += inc
    return result


def utc_now():
    """Current UTC date and time with the microsecond value normalized to zero."""
    return datetime.now(timezone.utc).replace(microsecond=0)


def localized_dt_string(dt, use_tz=None):
    """Convert datetime value to a string, localized for the specified timezone."""
    if not dt.tzinfo and not use_tz:
        return dt.strftime(DT_NAIVE)
    if not dt.tzinfo:
        return dt.replace(tzinfo=use_tz).strftime(DT_AWARE)
    return dt.astimezone(use_tz).strftime(DT_AWARE) if use_tz else dt.strftime(DT_AWARE)


def make_tzaware(dt, use_tz=None, localize=True):
    """Translate an aware datetime to a different timezone OR add timezone to naive datetime."""
    if not use_tz:
        use_tz = get_local_utcoffset()
    return dt.astimezone(use_tz) if localize else dt.replace(tzinfo=use_tz)


def get_local_utcoffset():
    """Get UTC offset from local system and return as timezone object."""
    utc_offset = timedelta(seconds=time.localtime().tm_gmtoff)
    return timezone(offset=utc_offset)


def dtaware_fromtimestamp(timestamp, use_tz=None):
    """Time-zone aware datetime object from UNIX timestamp."""
    timestamp_naive = datetime.fromtimestamp(timestamp)
    timestamp_aware = timestamp_naive.replace(tzinfo=get_local_utcoffset())
    return timestamp_aware.astimezone(use_tz) if use_tz else timestamp_aware


def today_str():
    return date.today().strftime(DATE_ONLY_2)


def current_year():
    return datetime.now().year


def format_timedelta_str(td, precise=True):
    """Convert timedelta to an easy-to-read string value."""
    (milliseconds, microseconds) = divmod(td.microseconds, 1000)
    (minutes, seconds) = divmod(td.seconds, 60)
    (hours, minutes) = divmod(minutes, 60)
    if td.days > 0:
        return f"{td.days}d {hours:.0f}h {minutes:.0f}m {seconds}s" if precise else f"{td.days} days"
    if hours > 0:
        return f"{hours:.0f}h {minutes:.0f}m {seconds}s" if precise else f"{hours:.0f} hours {minutes:.0f} minutes"
    if minutes > 0:
        return f"{minutes:.0f}m {seconds}s" if precise else f"{minutes:.0f} minutes"
    if td.seconds > 0:
        return f"{td.seconds}s {milliseconds:.0f}ms" if precise else f"{td.seconds} seconds"
    if milliseconds > 0:
        return f"{milliseconds}ms"
    return f"{microseconds}us"



from dataclasses import dataclass
from datetime import datetime

from dataclass_csv import accept_whitespaces, dateformat
from dateutil import tz
from sqlalchemy import Column, DateTime, ForeignKey, Integer, String
from sqlalchemy.ext.hybrid import hybrid_property
from sqlalchemy.orm import relationship

import vigorish.database as db
from vigorish.models.season import Season
from vigorish.util.dt_format_strings import (
    DATE_ONLY,
    DATE_ONLY_2,
    DATE_ONLY_TABLE_ID,
    DT_STR_FORMAT,
)
from vigorish.util.string_helpers import get_brooks_team_id, validate_brooks_game_id


class GameScrapeStatus(db.Base):
    __tablename__ = "scrape_status_game"
    id = Column(Integer, primary_key=True)
    game_date = Column(DateTime)
    game_time_hour = Column(Integer)
    game_time_minute = Column(Integer)
    game_time_zone = Column(String)
    bbref_game_id = Column(String, unique=True)
    bb_game_id = Column(String, unique=True)
    scraped_bbref_boxscore = Column(Integer, default=0)
    scraped_brooks_pitch_logs = Column(Integer, default=0)
    combined_data_success = Column(Integer, default=0)
    combined_data_fail = Column(Integer, default=0)
    imported_bat_stats = Column(Integer, default=0)
    imported_pitch_stats = Column(Integer, default=0)
    pitch_app_count_bbref = Column(Integer, default=0)
    pitch_app_count_brooks = Column(Integer, default=0)
    total_pitch_count_bbref = Column(Integer, default=0)
    scrape_status_date_id = Column(Integer, ForeignKey("scrape_status_date.id"))
    season_id = Column(Integer, ForeignKey("season.id"))

    pitch_apps = relationship("PitchAppScrapeStatus")
    pitchfx = relationship("PitchFx")

    @hybrid_property
    def game_start_time(self):
        game_start = datetime(
            year=self.game_date.year,
            month=self.game_date.month,
            day=self.game_date.day,
            hour=self.game_time_hour,
            minute=self.game_time_minute,
        )
        return (
            game_start.replace(tzinfo=tz.gettz(self.game_time_zone))
            if self.game_time_hour != 0 or self.game_time_minute != 0
            else None
        )

    @hybrid_property
    def game_date_time_str(self):
        return (
            self.game_start_time.strftime(DT_STR_FORMAT)
            if self.game_start_time
            else self.game_date.strftime(DATE_ONLY_2)
            if self.game_date
            else None
        )

    @hybrid_property
    def away_team_id_bb(self):
        game_dict = validate_brooks_game_id(self.bb_game_id).value
        return game_dict["away_team_id"]

    @hybrid_property
    def home_team_id_bb(self):
        game_dict = validate_brooks_game_id(self.bb_game_id).value
        return game_dict["home_team_id"]

    @hybrid_property
    def pitch_app_count_pitchfx(self):
        return len(self.pitch_apps)

    @hybrid_property
    def total_pitch_apps_scraped_pitchfx(self):
        return sum(pitch_app.scraped_pitchfx for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_pitch_apps_no_pitchfx_data(self):
        return sum(pitch_app.no_pitchfx_data for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_pitch_apps_with_pitchfx_data(self):
        return self.pitch_app_count_pitchfx - self.total_pitch_apps_no_pitchfx_data

    @hybrid_property
    def total_pitch_apps_combined_data(self):
        return sum(pitch_app.combined_pitchfx_bbref_data for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_pitch_apps_pitchfx_error(self):
        return sum(pitch_app.pitchfx_error for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_pitch_apps_invalid_pitchfx(self):
        return sum(pitch_app.invalid_pitchfx for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_pitch_apps_pitchfx_is_valid(self):
        return sum(not (pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.pitch_apps)

    @hybrid_property
    def total_pitch_count_pitch_logs(self):
        return sum(pitch_app.pitch_count_pitch_log for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_pitch_count_bbref_audited(self):
        return sum(pitch_app.pitch_count_bbref for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_pitch_count_pitchfx(self):
        return sum(pitch_app.pitch_count_pitchfx for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_pitch_count_pitchfx_audited(self):
        return sum(pitch_app.pitch_count_pitchfx_audited for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_missing_pitchfx_count(self):
        return sum(pitch_app.missing_pitchfx_count for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_removed_pitchfx_count(self):
        return sum(pitch_app.removed_pitchfx_count for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_batters_faced_bbref(self):
        return sum(pitch_app.batters_faced_bbref for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_batters_faced_pitchfx(self):
        return sum(pitch_app.batters_faced_pitchfx for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_at_bats_pitchfx_complete(self):
        return sum(pitch_app.total_at_bats_pitchfx_complete for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_at_bats_missing_pitchfx(self):
        return sum(pitch_app.total_at_bats_missing_pitchfx for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_at_bats_removed_pitchfx(self):
        return sum(pitch_app.total_at_bats_removed_pitchfx for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_at_bats_pitchfx_error(self):
        return sum(pitch_app.total_at_bats_pitchfx_error for pitch_app in self.pitch_apps)

    @hybrid_property
    def total_at_bats_invalid_pitchfx(self):
        return sum(pitch_app.total_at_bats_invalid_pitchfx for pitch_app in self.pitch_apps)

    @hybrid_property
    def scraped_all_pitchfx_logs(self):
        return (
            True
            if not self.pitch_apps
            else False
            if not self.scraped_brooks_pitch_logs
            else self.pitch_app_count_pitchfx == self.total_pitch_apps_scraped_pitchfx
        )

    @hybrid_property
    def combined_data_for_all_pitchfx_logs(self):
        return (
            False
            if not self.pitch_apps or not self.scraped_all_pitchfx_logs
            else self.pitch_app_count_pitchfx == self.total_pitch_apps_combined_data
        )

    @hybrid_property
    def pitchfx_error_for_any_pitchfx_logs(self):
        return (
            False
            if not self.pitch_apps or not self.scraped_all_pitchfx_logs
            else any((pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.pitch_apps)
        )

    @hybrid_property
    def pitchfx_is_valid_for_all_pitchfx_logs(self):
        return (
            False
            if not self.pitch_apps or not self.scraped_all_pitchfx_logs
            else all(not (pfx.pitchfx_error or pfx.invalid_pitchfx) for pfx in self.pitch_apps)
        )

    def __repr__(self):
        return f"<GameScrapeStatus bbref_game_id={self.bbref_game_id}>"

    def as_dict(self):
        return {
            "bbref_game_id": self.bbref_game_id,
            "bb_game_id": self.bb_game_id,
            "game_date_time_str": self.game_date_time_str,
            "scraped_bbref_boxscore": self.scraped_bbref_boxscore,
            "scraped_brooks_pitch_logs": self.scraped_brooks_pitch_logs,
            "combined_data_success": self.combined_data_success,
            "combined_data_fail": self.combined_data_fail,
            "pitch_app_count_bbref": self.pitch_app_count_bbref,
            "pitch_app_count_brooks": self.pitch_app_count_brooks,
            "pitch_app_count_pitchfx": self.pitch_app_count_pitchfx,
            "total_pitch_apps_scraped_pitchfx": self.total_pitch_apps_scraped_pitchfx,
            "total_pitch_apps_no_pitchfx_data": self.total_pitch_apps_no_pitchfx_data,
            "total_pitch_apps_with_pitchfx_data": self.total_pitch_apps_with_pitchfx_data,
            "total_pitch_apps_combined_data": self.total_pitch_apps_combined_data,
            "total_pitch_apps_pitchfx_error": self.total_pitch_apps_pitchfx_error,
            "total_pitch_apps_invalid_pitchfx": self.total_pitch_apps_invalid_pitchfx,
            "total_pitch_apps_pitchfx_is_valid": self.total_pitch_apps_pitchfx_is_valid,
            "total_pitch_count_pitch_logs": self.total_pitch_count_pitch_logs,
            "total_pitch_count_bbref": self.total_pitch_count_bbref,
            "total_pitch_count_bbref_audited": self.total_pitch_count_bbref_audited,
            "total_pitch_count_pitchfx": self.total_pitch_count_pitchfx,
            "total_pitch_count_pitchfx_audited": self.total_pitch_count_pitchfx_audited,
            "total_missing_pitchfx_count": self.total_missing_pitchfx_count,
            "total_removed_pitchfx_count": self.total_removed_pitchfx_count,
            "total_batters_faced_bbref": self.total_batters_faced_bbref,
            "total_batters_faced_pitchfx": self.total_batters_faced_pitchfx,
            "total_at_bats_pitchfx_complete": self.total_at_bats_pitchfx_complete,
            "total_at_bats_missing_pitchfx": self.total_at_bats_missing_pitchfx,
            "total_at_bats_removed_pitchfx": self.total_at_bats_removed_pitchfx,
            "total_at_bats_pitchfx_error": self.total_at_bats_pitchfx_error,
            "total_at_bats_invalid_pitchfx": self.total_at_bats_invalid_pitchfx,
            "scraped_all_pitchfx_logs": self.scraped_all_pitchfx_logs,
            "combined_data_for_all_pitchfx_logs": self.combined_data_for_all_pitchfx_logs,
            "pitchfx_error_for_any_pitchfx_logs": self.pitchfx_error_for_any_pitchfx_logs,
            "pitchfx_is_valid_for_all_pitchfx_logs": self.pitchfx_is_valid_for_all_pitchfx_logs,
        }

    def status_report(self):
        bbref_game_id = self.bbref_game_id or "BBREF_GAME_ID IS MISSING!"
        bb_game_id = self.bb_game_id or "BB_GAME_ID IS MISSING!"
        scraped_bbref_boxscore = "YES" if self.scraped_bbref_boxscore == 1 else "NO"
        scraped_brooks_pitch_logs = "YES" if self.scraped_brooks_pitch_logs == 1 else "NO"
        scraped_all_pitchfx_logs = "YES" if self.scraped_all_pitchfx_logs else "NO"
        combined_data_for_all_pitchfx_logs = "YES" if self.combined_data_for_all_pitchfx_logs else "NO"
        pitchfx_error_for_any_pitchfx_logs = "YES" if self.pitchfx_error_for_any_pitchfx_logs else "NO"
        return [
            f"BBRef Game ID................................: {bbref_game_id}",
            f"brooksbaseball.net Game ID...................: {bb_game_id}",
            f"Game Date-Time...............................: {self.game_date_time_str}",
            f"Scraped BBRef Boxscore.......................: {scraped_bbref_boxscore}",
            f"Scraped Brooks Pitch Logs....................: {scraped_brooks_pitch_logs}",
            (
                f"PitchFx Logs Scraped.........................: {scraped_all_pitchfx_logs} "
                f"{self.total_pitch_apps_scraped_pitchfx}/{self.pitch_app_count_pitchfx}"
            ),
            f"Combined BBRef/PitchFX Data..................: {combined_data_for_all_pitchfx_logs}",
            (
                f"PitchFX Data Errors (Valid AB/Invalid AB)....: "
                f"{pitchfx_error_for_any_pitchfx_logs} "
                f"{self.total_pitch_apps_pitchfx_error}/{self.total_pitch_apps_invalid_pitchfx}"
            ),
            (
                f"Pitch App Count (BBRef/Brooks)...............: "
                f"{self.pitch_app_count_bbref}/{self.pitch_app_count_brooks}"
            ),
            (
                f"Pitch App Count (PFx/data/no data)...........: "
                f"{self.pitch_app_count_pitchfx}/"
                f"{self.total_pitch_apps_with_pitchfx_data}/"
                f"{self.total_pitch_apps_no_pitchfx_data}"
            ),
            (
                f"Pitch Count (BBRef/Brooks/PFx)...............: {self.total_pitch_count_bbref}/"
                f"{self.total_pitch_count_pitch_logs}/{self.total_pitch_count_pitchfx}"
            ),
            (
                "Pitch Count Audited (BBRef/PFx/Removed)......: "
                f"{self.total_pitch_count_bbref_audited}/"
                f"{self.total_pitch_count_pitchfx_audited}/"
                f"{self.total_removed_pitchfx_count}"
            ),
        ]

    @classmethod
    def find_by_bbref_game_id(cls, db_session, bbref_game_id):
        return db_session.query(cls).filter_by(bbref_game_id=bbref_game_id).first()

    @classmethod
    def find_by_bb_game_id(cls, db_session, bb_game_id):
        return db_session.query(cls).filter_by(bb_game_id=bb_game_id).first()

    @classmethod
    def get_all_scraped_bbref_game_ids_for_season(cls, db_session, season_id):
        return [
            game_status.bbref_game_id
            for game_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if game_status.scraped_bbref_boxscore == 1
        ]

    @classmethod
    def get_all_scraped_brooks_game_ids_for_season(cls, db_session, season_id):
        return [
            game_status.bb_game_id
            for game_status in db_session.query(cls).filter_by(season_id=season_id).all()
            if game_status.scraped_brooks_pitch_logs == 1
        ]

    @classmethod
    def get_all_bbref_game_ids(cls, db_session, season_id):
        return [game_status.bbref_game_id for game_status in db_session.query(cls).filter_by(season_id=season_id).all()]

    @classmethod
    def get_all_brooks_game_ids_for_date(cls, db_session, date):
        date_id = date.strftime(DATE_ONLY_TABLE_ID)
        return [
            game.bb_game_id
            for game in db_session.query(cls)
            .filter_by(scrape_status_date_id=int(date_id))
            .order_by(GameScrapeStatus.bb_game_id)
            .all()
        ]

    @classmethod
    def get_all_bbref_game_ids_for_team(cls, db_session, team_id_br, year):
        season = Season.find_by_year(db_session, year)
        if not season:
            return None
        games_for_season = list(db_session.query(cls).filter_by(season_id=season.id).all())
        games_for_season.sort(key=lambda x: x.scrape_status_date_id)
        bb_team_id = get_brooks_team_id(team_id_br)
        return [
            game.bbref_game_id
            for game in games_for_season
            if game.away_team_id_bb == bb_team_id or game.home_team_id_bb == bb_team_id
        ]

    @classmethod
    def get_game_id_map(cls, db_session):
        all_games = db_session.query(cls).all()
        return {g.bbref_game_id: g.id for g in all_games}


@accept_whitespaces
@dateformat(DATE_ONLY)
@dataclass
class GameScrapeStatusCsvRow:
    id: int
    game_date: datetime = None
    game_time_hour: int = 0
    game_time_minute: int = 0
    game_time_zone: str = None
    bbref_game_id: str = None
    bb_game_id: str = None
    scraped_bbref_boxscore: int = 0
    scraped_brooks_pitch_logs: int = 0
    combined_data_success: int = 0
    combined_data_fail: int = 0
    imported_bat_stats: int = 0
    imported_pitch_stats: int = 0
    pitch_app_count_bbref: int = 0
    pitch_app_count_brooks: int = 0
    total_pitch_count_bbref: int = 0



"""Populate player_id and player tables with initial data."""
from dataclasses import dataclass
from datetime import datetime

from dataclass_csv import accept_whitespaces, DataclassReader, dateformat
from tqdm import tqdm

import vigorish.database as db
import vigorish.tasks as tasks
from vigorish.util.dt_format_strings import DATE_ONLY
from vigorish.util.result import Result


@accept_whitespaces
@dateformat(DATE_ONLY)
@dataclass
class PlayerCsvRow:
    playerID: str = None
    birthYear: int = None
    birthMonth: int = None
    birthDay: int = None
    birthCountry: str = None
    birthState: str = None
    birthCity: str = None
    deathYear: int = None
    deathMonth: int = None
    deathDay: int = None
    deathCountry: str = None
    deathState: str = None
    deathCity: str = None
    nameFirst: str = None
    nameLast: str = None
    nameGiven: str = None
    weight: int = None
    height: int = None
    bats: str = None
    throws: str = None
    debut: datetime = None
    finalGame: datetime = None
    retroID: str = None
    bbrefID: str = None


PLAYER_ID_MAP_CSV = "bbref_player_id_map.csv"
PLAYER_TEAM_MAP_CSV = "bbref_player_team_map.csv"
PEOPLE_CSV = "People.csv"


def populate_players(app, csv_folder):
    """Populate player_id and player tables with initial data."""
    result = import_id_map_csv(app, csv_folder)
    if result.failure:
        return result
    app.db_session.commit()
    result = import_people_csv(app, csv_folder)
    if result.failure:
        return result
    app.db_session.commit()
    result = import_team_map_csv(app, csv_folder)
    if result.failure:
        return result
    app.db_session.commit()
    return Result.Ok()


def import_id_map_csv(app, csv_folder):
    try:
        id_map_task = tasks.UpdatePlayerIdMapTask(app, csv_folder.joinpath(PLAYER_ID_MAP_CSV))
        player_id_map = id_map_task.read_bbref_player_id_map_from_file()
        with tqdm(
            total=len(player_id_map),
            desc="Populating player_id table.....",
            unit="row",
            mininterval=0.12,
            maxinterval=5,
            unit_scale=True,
            ncols=90,
        ) as pbar:
            for id_map in player_id_map:
                app.db_session.add(
                    db.PlayerId(
                        mlb_id=int(id_map.mlb_ID),
                        mlb_name=id_map.name_common,
                        bbref_id=id_map.player_ID,
                        bbref_name=None,
                    )
                )
                pbar.update()
        return Result.Ok()
    except Exception as e:
        error = f"Error: {repr(e)}"
        app.db_session.rollback()
        return Result.Fail(error)


def import_people_csv(app, csv_folder):
    player_csv_file = csv_folder.joinpath(PEOPLE_CSV)
    csv_text = player_csv_file.read_text()
    total_rows = len([row for row in csv_text.split("\n") if row])
    # To avoid populating the database with players from the 19th century, we set a limit based on the
    # year a player debuted to determine if they need to be added. First, the earliest year (season)
    # from the season table is found. Since two players have tied for the longest MLB career with 27
    # years, setting a limit of 30 years will ensure that only players that could potentially have
    # data tracked by this database are added.
    earliest_season = min(s.year for s in db.Season.get_all_regular_seasons(app.db_session))
    player_debut_limit = earliest_season - 30
    try:
        with open(player_csv_file) as player_csv:
            reader = DataclassReader(player_csv, PlayerCsvRow)
            with tqdm(
                total=total_rows,
                desc="Populating player table........",
                unit="row",
                mininterval=0.12,
                maxinterval=5,
                unit_scale=True,
                ncols=90,
            ) as pbar:
                for row in reader:
                    if not (row.birthYear or row.birthMonth or row.birthDay or row.debut):
                        pbar.update()
                        continue
                    if row.debut and row.debut.year < player_debut_limit:
                        pbar.update()
                        continue
                    player_id = db.PlayerId.find_by_bbref_id(app.db_session, row.bbrefID)
                    if not player_id:
                        pbar.update()
                        continue
                    p = db.Player(
                        mlb_id=player_id.mlb_id,
                        bbref_id=row.bbrefID,
                        name_first=row.nameFirst,
                        name_last=row.nameLast,
                        name_given=row.nameGiven,
                        bats=row.bats,
                        throws=row.throws,
                        weight=row.weight,
                        height=row.height,
                        debut=row.debut,
                        birth_year=row.birthYear,
                        birth_month=row.birthMonth,
                        birth_day=row.birthDay,
                        birth_country=row.birthCountry,
                        birth_state=row.birthState,
                        birth_city=row.birthCity,
                        missing_mlb_id=False,
                    )
                    app.db_session.add(p)
                    app.db_session.commit()
                    player_id.db_player_id = p.id
                    pbar.update()
        return Result.Ok()
    except Exception as e:
        error = f"Error: {repr(e)}"
        app.db_session.rollback()
        return Result.Fail(error)


def import_team_map_csv(app, csv_folder):
    try:
        team_map_task = tasks.UpdatePlayerTeamMapTask(app, csv_folder.joinpath(PLAYER_TEAM_MAP_CSV))
        player_team_map = team_map_task.read_bbref_player_team_map_from_file()
        with tqdm(
            total=len(player_team_map),
            desc="Populating player_team table...",
            unit="row",
            mininterval=0.12,
            maxinterval=5,
            unit_scale=True,
            ncols=90,
        ) as pbar:
            for team_map in player_team_map:
                player = db.Player.find_by_bbref_id(app.db_session, team_map.player_ID)
                if not player:
                    pbar.update()
                    continue
                team = db.Team.find_by_team_id_and_year(app.db_session, team_map.team_ID, int(team_map.year_ID))
                season = db.Season.find_by_year(app.db_session, int(team_map.year_ID))
                player_team = db.Assoc_Player_Team(
                    db_player_id=player.id,
                    db_team_id=team.id,
                    mlb_id=int(team_map.mlb_ID),
                    bbref_id=team_map.player_ID,
                    team_id=team.team_id_br,
                    year=int(team_map.year_ID),
                    stint_number=team_map.stint_ID,
                    season_id=season.id,
                )
                app.db_session.add(player_team)
                pbar.update()
        return Result.Ok()
    except Exception as e:
        error = f"Error: {repr(e)}"
        app.db_session.rollback()
        return Result.Fail(error)



# flake8: noqa
from vigorish.tasks.add_to_database import AddToDatabaseTask
from vigorish.tasks.backup_database import BackupDatabaseTask
from vigorish.tasks.calculate_avg_pitch_times import CalculateAvgPitchTimesTask
from vigorish.tasks.combine_scraped_data import CombineScrapedDataTask
from vigorish.tasks.import_scraped_data import ImportScrapedDataTask
from vigorish.tasks.patch_all_invalid_pfx import PatchAllInvalidPitchFxTask
from vigorish.tasks.restore_database import RestoreDatabaseTask
from vigorish.tasks.sync_data_no_prompts import SyncDataNoPromptsTask
from vigorish.tasks.update_player_maps import UpdatePlayerIdMapTask, UpdatePlayerTeamMapTask



# flake8: noqa
from vigorish.data.metrics.bat_stats import BatStatsMetrics
from vigorish.data.metrics.pitch_stats import PitchStatsMetrics
from vigorish.data.metrics.pitchfx import PitchFxMetrics, PitchFxMetricsCollection



"""Populate mlb_season table with initial data."""
from datetime import date

from tqdm import tqdm

import vigorish.database as db
from vigorish.enums import SeasonType
from vigorish.util.result import Result


def populate_seasons(db_session):
    """Populate mlb_season table with initial data."""
    result = add_mlb_seasons(db_session)
    if result.failure:
        return result
    db_session.commit()
    return Result.Ok()


def add_mlb_seasons(db_session):
    try:
        mlb_seasons = [
            db.Season(
                year=2014,
                start_date=date(2014, 3, 22),
                end_date=date(2014, 9, 28),
                asg_date=date(2014, 7, 14),
                season_type=SeasonType.REGULAR_SEASON,
            ),
            db.Season(
                year=2014,
                start_date=date(2014, 9, 30),
                end_date=date(2014, 10, 29),
                season_type=SeasonType.POST_SEASON,
            ),
            db.Season(
                year=2015,
                start_date=date(2015, 4, 5),
                end_date=date(2015, 10, 4),
                asg_date=date(2015, 7, 14),
                season_type=SeasonType.REGULAR_SEASON,
            ),
            db.Season(
                year=2015,
                start_date=date(2015, 10, 6),
                end_date=date(2015, 11, 1),
                season_type=SeasonType.POST_SEASON,
            ),
            db.Season(
                year=2016,
                start_date=date(2016, 4, 3),
                end_date=date(2016, 10, 2),
                asg_date=date(2016, 7, 12),
                season_type=SeasonType.REGULAR_SEASON,
            ),
            db.Season(
                year=2016,
                start_date=date(2016, 10, 3),
                end_date=date(2016, 11, 2),
                season_type=SeasonType.POST_SEASON,
            ),
            db.Season(
                year=2017,
                start_date=date(2017, 4, 2),
                end_date=date(2017, 10, 1),
                asg_date=date(2017, 7, 11),
                season_type=SeasonType.REGULAR_SEASON,
            ),
            db.Season(
                year=2017,
                start_date=date(2017, 10, 3),
                end_date=date(2017, 11, 1),
                season_type=SeasonType.POST_SEASON,
            ),
            db.Season(
                year=2018,
                start_date=date(2018, 3, 29),
                end_date=date(2018, 10, 1),
                asg_date=date(2018, 7, 17),
                season_type=SeasonType.REGULAR_SEASON,
            ),
            db.Season(
                year=2018,
                start_date=date(2018, 10, 2),
                end_date=date(2018, 10, 31),
                season_type=SeasonType.POST_SEASON,
            ),
            db.Season(
                year=2019,
                start_date=date(2019, 3, 28),
                end_date=date(2019, 9, 29),
                asg_date=date(2019, 7, 9),
                season_type=SeasonType.REGULAR_SEASON,
            ),
            db.Season(
                year=2019,
                start_date=date(2019, 10, 1),
                end_date=date(2019, 10, 30),
                season_type=SeasonType.POST_SEASON,
            ),
            db.Season(
                year=2020,
                start_date=date(2020, 7, 23),
                end_date=date(2020, 9, 27),
                season_type=SeasonType.REGULAR_SEASON,
            ),
            db.Season(
                year=2020,
                start_date=date(2020, 9, 29),
                end_date=date(2020, 10, 31),
                season_type=SeasonType.POST_SEASON,
            ),
        ]

        for season in tqdm(
            mlb_seasons,
            desc="Populating mlb_season table....",
            unit="row",
            mininterval=0.12,
            maxinterval=5,
            unit_scale=True,
            ncols=90,
        ):
            db_session.add(season)
        return Result.Ok()
    except Exception as e:
        error = f"Error: {repr(e)}"
        db_session.rollback()
        return Result.Fail(error)



from __future__ import annotations

from typing import Callable, Generic, Iterable, Optional, TypeVar

T = TypeVar("T")


class Result(Generic[T]):
    """Represent the outcome of an operation."""

    def __init__(self, success: bool, value: T, error: str) -> None:
        self.success = success
        self.error = error
        self.value = value

    def __str__(self) -> str:
        """Informal string representation of a result."""
        result = "Success" if self.success else "Fail"
        detail = f" {self.error}" if self.failure else f" value={self.value}" if self.value else ""
        return f"[{result}]{detail}"

    def __repr__(self) -> str:
        """Official string representation of a result."""
        detail = f', error="{self.error}"' if self.failure else f", value={self.value}" if self.value else ""
        return f"<Result success={self.success}{detail}>"

    @property
    def failure(self) -> bool:
        """Flag that indicates if the operation failed."""
        return not self.success

    def on_success(self, func: Callable, *args, **kwargs) -> Result:
        """Pass result of successful operation (if any) to subsequent function."""
        return self if self.failure else func(self.value, *args, **kwargs) if self.value else func(*args, **kwargs)

    def on_failure(self, func: Callable, *args, **kwargs) -> Result:
        """Pass error message from failed operation to subsequent function."""
        return self if self.success else func(self.error, *args, **kwargs)

    def on_both(self, func: Callable, *args, **kwargs) -> Result:
        """Pass result (either succeeded/failed) to subsequent function."""
        return func(self, *args, **kwargs)

    @staticmethod
    def Fail(error_message: str) -> Result:
        """Create a Result object for a failed operation."""
        return Result(False, value=None, error=error_message)

    @staticmethod
    def Ok(value: Optional[T] = None) -> Result[T]:
        """Create a Result object for a successful operation."""
        return Result(True, value=value, error=None)

    @staticmethod
    def Combine(results: Iterable[Result]) -> Result:
        """Return a Result object based on the outcome of a list of Results."""
        return (
            Result.Ok([result.value or None for result in results])
            if all(result.success for result in results)
            else Result.Fail([result.error if result.failure else None for result in results])
        )
